import unittest
import uuid
from unittest import mock

import colander
from pyramid.response import Response

from kinto.core.testing import DummyRequest
from kinto.core.utils import json
from kinto.core.views.batch import BatchPayloadSchema
from kinto.core.views.batch import batch as batch_service

from .support import BaseWebTest


class BatchViewTest(BaseWebTest, unittest.TestCase):
    def test_does_not_require_authentication(self):
        body = {"requests": []}
        self.app.post_json("/batch", body)

    def test_returns_400_if_body_has_missing_requests(self):
        self.app.post("/batch", {}, headers=self.headers, status=400)

    def test_returns_responses_if_schema_is_valid(self):
        body = {"requests": []}
        resp = self.app.post_json("/batch", body, headers=self.headers)
        self.assertIn("responses", resp.json)

    def test_defaults_are_applied_to_requests(self):
        request = {"path": "/v0/"}
        defaults = {"method": "POST"}
        result = self.app.post_json("/batch", {"requests": [request], "defaults": defaults})
        self.assertEqual(result.json["responses"][0]["status"], 405)

    def test_only_post_is_allowed(self):
        self.app.get("/batch", headers=self.headers, status=405)
        self.app.put("/batch", headers=self.headers, status=405)
        self.app.patch("/batch", headers=self.headers, status=405)
        self.app.delete("/batch", headers=self.headers, status=405)

    def test_batch_adds_missing_api_with_prefix(self):
        request = {"path": "/v0/"}
        body = {"requests": [request]}
        resp = self.app.post_json("/batch", body, headers=self.headers)
        hello = resp.json["responses"][0]
        self.assertEqual(hello["path"], "/v0/")
        self.assertEqual(hello["status"], 200)
        self.assertEqual(hello["body"]["project_name"], "myapp")
        self.assertIn("application/json", hello["headers"]["Content-Type"])

    def test_empty_response_body_with_head(self):
        request = {"path": "/v0/", "method": "HEAD"}
        body = {"requests": [request]}
        resp = self.app.post_json("/batch", body, headers=self.headers)
        head = resp.json["responses"][0]
        self.assertEqual(head["body"], "")
        self.assertNotEqual(len(head["headers"]), 0)

    def test_api_errors_are_json_formatted(self):
        request = {"path": "/unknown"}
        body = {"requests": [request]}
        resp = self.app.post_json("/batch", body, headers=self.headers)
        error = resp.json["responses"][0]
        self.assertEqual(error["body"]["code"], 404)

    def test_internal_errors_makes_the_batch_fail(self):
        request = {"path": "/v0/"}
        body = {"requests": [request]}

        with mock.patch("kinto.core.views.hello.get_eos") as mocked:
            mocked.side_effect = AttributeError
            self.app.post_json("/batch", body, headers=self.headers, status=500)

    def test_errors_handled_by_view_does_not_make_the_batch_fail(self):
        from requests.exceptions import HTTPError

        request = {"path": "/v0/"}
        body = {"requests": [request]}

        with mock.patch("kinto.core.views.hello.get_eos") as mocked:
            response = mock.MagicMock(status_code=404)
            mocked.side_effect = HTTPError(response=response)
            resp = self.app.post_json("/batch", body, headers=self.headers, status=200)
            subresponse = resp.json["responses"][0]["body"]
            self.assertEqual(subresponse, {"errno": 999, "code": 404, "error": "Not Found"})

    def test_batch_cannot_be_recursive(self):
        requests = {"requests": [{"path": "/v0/"}]}
        request = {"method": "POST", "path": "/v0/batch", "body": requests}
        body = {"requests": [request]}
        resp = self.app.post_json("/batch", body, status=400)
        self.assertIn("Recursive", resp.json["message"])

    def test_batch_validates_json(self):
        body = """{"requests": [{"path": "/v0/"},]}"""
        resp = self.app.post(
            "/batch", body, status=400, headers={"Content-Type": "application/json"}
        )
        self.assertIn("Invalid JSON", resp.json["message"])

    def test_batch_should_reject_unaccepted_request_content_type(self):
        request = {"requests": [{"path": "/v0/mushrooms"}]}
        self.app.post("/batch", request, status=415, headers={"Content-Type": "text/plain"})

    def test_responses_are_resolved_with_api_with_prefix(self):
        request = {"path": "/"}
        body = {"requests": [request]}
        resp = self.app.post_json("/batch", body, headers=self.headers)
        hello = resp.json["responses"][0]
        self.assertEqual(hello["path"], "/v0/")
        self.assertEqual(hello["status"], 200)
        self.assertEqual(hello["body"]["project_name"], "myapp")
        self.assertIn("application/json", hello["headers"]["Content-Type"])

    def test_redirect_responses_are_followed(self):
        request = {"path": "/mushrooms/"}  # trailing slash
        body = {"requests": [request]}
        resp = self.app.post_json("/batch", body, headers=self.headers)
        resource = resp.json["responses"][0]
        self.assertEqual(resource["status"], 200)
        self.assertEqual(resource["path"], "/v0/mushrooms")
        self.assertEqual(resource["body"], {"data": []})

    def test_body_is_transmitted_during_redirect(self):
        request = {
            "method": "PUT",
            "path": "/mushrooms/{}/".format(str(uuid.uuid4())),
            "body": {"data": {"name": "Trompette de la mort"}},
        }
        body = {"requests": [request]}
        resp = self.app.post_json("/batch", body, headers=self.headers)
        response = resp.json["responses"][0]
        self.assertEqual(response["status"], 201)
        obj = response["body"]["data"]
        self.assertEqual(obj["name"], "Trompette de la mort")

    def test_400_error_message_is_forwarded(self):
        headers = {**self.headers, "If-Match": '"*"'}
        request = {
            "method": "PUT",
            "path": "/mushrooms/{}".format(str(uuid.uuid4())),
            "body": {"data": {"name": "Trompette de la mort"}},
            "headers": headers,
        }
        body = {"requests": [request, request]}
        resp = self.app.post_json("/batch", body, status=200)
        self.assertEqual(resp.json["responses"][1]["status"], 400)
        msg = "If-Match in header: The value should be integer between double quotes."
        self.assertEqual(resp.json["responses"][1]["body"]["message"], msg)

    def test_412_errors_are_forwarded(self):
        headers = {**self.headers, "If-None-Match": "*"}
        request = {
            "method": "PUT",
            "path": "/mushrooms/{}".format(str(uuid.uuid4())),
            "body": {"data": {"name": "Trompette de la mort"}},
            "headers": headers,
        }
        body = {"requests": [request, request]}
        resp = self.app.post_json("/batch", body, status=200)
        self.assertEqual(resp.json["responses"][0]["status"], 201)
        self.assertEqual(resp.json["responses"][1]["status"], 412)


class BatchSchemaTest(unittest.TestCase):
    def setUp(self):
        self.schema = BatchPayloadSchema()

    def assertInvalid(self, payload):
        self.assertRaises(colander.Invalid, self.schema.deserialize, payload)

    def test_requests_is_mandatory(self):
        self.assertInvalid({})

    def test_raise_invalid_on_unknown_attributes(self):
        self.assertInvalid({"requests": [], "unknown": 42})

    def test_list_of_requests_can_be_empty(self):
        self.schema.deserialize({"requests": []})

    def test_list_of_requests_must_be_a_list(self):
        self.assertInvalid({"requests": {}})

    def test_list_of_requests_must_be_dicts(self):
        request = 42
        self.assertInvalid({"defaults": {"path": "/"}, "requests": [request]})

    def test_request_path_must_start_with_slash(self):
        request = {"path": "http://localhost"}
        self.assertInvalid({"requests": [request]})

    def test_request_path_is_mandatory(self):
        request = {"method": "HEAD"}
        self.assertInvalid({"requests": [request]})

    def test_request_method_must_be_known_uppercase_word(self):
        request = {"path": "/", "method": "get"}
        self.assertInvalid({"requests": [request]})

    def test_raise_invalid_on_request_unknown_attributes(self):
        request = {"path": "/", "method": "GET", "foo": 42}
        self.assertInvalid({"requests": [request]})

    #
    # headers
    #

    def test_request_headers_should_be_strings(self):
        headers = {"Accept": 3.14}
        request = {"path": "/", "headers": headers}
        self.assertInvalid({"requests": [request]})

    def test_request_headers_cannot_be_recursive(self):
        headers = {"Accept": {"sub": "dict"}}
        request = {"path": "/", "headers": headers}
        self.assertInvalid({"requests": [request]})

    def test_request_headers_are_preserved(self):
        headers = {"Accept": "audio/*"}
        request = {"path": "/", "headers": headers}
        deserialized = self.schema.deserialize({"requests": [request]})
        self.assertEqual(deserialized["requests"][0]["headers"]["Accept"], "audio/*")

    #
    # body
    #

    def test_body_is_an_arbitrary_mapping(self):
        payload = {"json": "payload"}
        request = {"path": "/", "body": payload}
        deserialized = self.schema.deserialize({"requests": [request]})
        self.assertEqual(deserialized["requests"][0]["body"], payload)

    #
    # defaults
    #

    def test_defaults_must_be_a_mapping_if_specified(self):
        request = {"path": "/"}
        batch_payload = {"requests": [request], "defaults": 42}
        self.assertInvalid(batch_payload)

    def test_defaults_must_be_a_request_schema_if_specified(self):
        request = {"path": "/"}
        defaults = {"body": 3}
        batch_payload = {"requests": [request], "defaults": defaults}
        self.assertInvalid(batch_payload)

    def test_raise_invalid_on_default_unknown_attributes(self):
        request = {"path": "/"}
        defaults = {"foo": "bar"}
        self.assertInvalid({"requests": [request], "defaults": defaults})

    def test_defaults_can_be_specified_empty(self):
        request = {"path": "/"}
        defaults = {}
        batch_payload = {"requests": [request], "defaults": defaults}
        self.schema.deserialize(batch_payload)

    def test_defaults_path_is_applied_to_requests(self):
        request = {"method": "GET"}
        defaults = {"path": "/"}
        batch_payload = {"requests": [request], "defaults": defaults}
        result = self.schema.deserialize(batch_payload)
        self.assertEqual(result["requests"][0]["path"], "/")

    def test_defaults_body_is_applied_to_requests(self):
        request = {"path": "/"}
        defaults = {"body": {"json": "payload"}}
        batch_payload = {"requests": [request], "defaults": defaults}
        result = self.schema.deserialize(batch_payload)
        self.assertEqual(result["requests"][0]["body"], {"json": "payload"})

    def test_defaults_headers_are_applied_to_requests(self):
        request = {"path": "/"}
        defaults = {"headers": {"Content-Type": "text/html"}}
        batch_payload = {"requests": [request], "defaults": defaults}
        result = self.schema.deserialize(batch_payload)
        self.assertEqual(result["requests"][0]["headers"]["Content-Type"], "text/html")

    def test_defaults_values_do_not_overwrite_requests_values(self):
        request = {"path": "/", "headers": {"Authorization": "me"}}
        defaults = {"headers": {"Authorization": "you", "Accept": "*/*"}}
        batch_payload = {"requests": [request], "defaults": defaults}
        result = self.schema.deserialize(batch_payload)
        self.assertEqual(
            result["requests"][0]["headers"], {"Authorization": "me", "Accept": "*/*"}
        )

    def test_defaults_values_for_path_must_start_with_slash(self):
        request = {}
        defaults = {"path": "http://localhost"}
        batch_payload = {"requests": [request], "defaults": defaults}
        self.assertInvalid(batch_payload)


class BatchServiceTest(unittest.TestCase):
    def setUp(self):
        self.method, self.view, self.options = batch_service.definitions[0]
        self.request = DummyRequest()

    def post(self, validated):
        self.request.validated = {"body": validated}
        return self.view(self.request)

    def test_returns_empty_list_of_responses_if_requests_empty(self):
        result = self.post({"requests": []})
        self.assertEqual(result["responses"], [])

    def test_returns_one_response_per_request(self):
        requests = [{"path": "/"}]
        result = self.post({"requests": requests})
        self.assertEqual(len(result["responses"]), len(requests))

    def test_relies_on_pyramid_invoke_subrequest(self):
        self.post({"requests": [{"path": "/"}]})
        self.assertTrue(self.request.invoke_subrequest.called)

    def test_returns_requests_path_in_responses(self):
        result = self.post({"requests": [{"path": "/"}]})
        self.assertEqual(result["responses"][0]["path"], "/v0/")

    def test_subrequests_have_parent_attribute(self):
        self.request.path = "/batch"
        self.post({"requests": [{"path": "/"}]})
        (subrequest,) = self.request.invoke_subrequest.call_args[0]
        self.assertEqual(subrequest.parent.path, "/batch")

    def test_subrequests_are_GET_by_default(self):
        self.post({"requests": [{"path": "/"}]})
        (subrequest,) = self.request.invoke_subrequest.call_args[0]
        self.assertEqual(subrequest.method, "GET")

    def test_original_request_headers_are_passed_to_subrequests(self):
        self.request.headers["Authorization"] = "Basic ertyfghjkl"
        self.post({"requests": [{"path": "/"}]})
        (subrequest,) = self.request.invoke_subrequest.call_args[0]
        self.assertIn("Basic", subrequest.headers["Authorization"])

    def test_subrequests_body_are_json_serialized(self):
        request = {"path": "/", "body": {"json": "payload"}}
        self.post({"requests": [request]})
        wanted = {"json": "payload"}
        (subrequest,) = self.request.invoke_subrequest.call_args[0]
        self.assertEqual(subrequest.body.decode("utf8"), json.dumps(wanted))

    def test_subrequests_body_have_json_content_type(self):
        self.request.headers["Content-Type"] = "text/xml"
        request = {"path": "/", "body": {"json": "payload"}}
        self.post({"requests": [request]})
        (subrequest,) = self.request.invoke_subrequest.call_args[0]
        self.assertIn("application/json", subrequest.headers["Content-Type"])

    def test_subrequests_body_have_utf8_charset(self):
        request = {"path": "/", "body": {"json": "😂"}}
        self.post({"requests": [request]})
        (subrequest,) = self.request.invoke_subrequest.call_args[0]
        self.assertIn("charset=utf-8", subrequest.headers["Content-Type"])
        wanted = {"json": "😂"}
        self.assertEqual(subrequest.body.decode("utf8"), json.dumps(wanted))

    def test_subrequests_paths_are_url_encoded(self):
        request = {"path": "/test?param=©"}
        self.post({"requests": [request]})
        (subrequest,) = self.request.invoke_subrequest.call_args[0]
        self.assertEqual(subrequest.path, "/v0/test")
        self.assertEqual(subrequest.GET["param"], "©")

    def test_subrequests_responses_paths_are_url_decoded(self):
        request = {"path": "/test?param=©"}
        resp = self.post({"requests": [request]})
        path = resp["responses"][0]["path"]
        self.assertEqual(path, "/v0/test")

    def test_response_body_is_string_if_remote_response_is_not_json(self):
        response = Response(body="Internal Error")
        self.request.invoke_subrequest.return_value = response
        request = {"path": "/test"}
        resp = self.post({"requests": [request]})
        body = resp["responses"][0]["body"].decode("utf-8")
        self.assertEqual(body, "Internal Error")

    def test_number_of_requests_is_not_limited_when_settings_set_to_none(self):
        self.request.registry.settings["batch_max_requests"] = None
        requests = {}
        for i in range(30):
            requests.setdefault("requests", []).append({"path": "/"})
            self.post(requests)

    def test_number_of_requests_is_limited_to_25_by_default(self):
        requests = {}
        for i in range(26):
            requests.setdefault("requests", []).append({"path": "/"})
        result = self.post(requests)

        self.assertEqual(
            self.request.errors[0]["description"], "Number of requests is limited to 25"
        )
        self.assertIsNone(result)  # rest of view not executed

    def test_return_400_if_number_of_requests_is_greater_than_settings(self):
        self.request.registry.settings["batch_max_requests"] = 22

        requests = {}
        for i in range(23):
            requests.setdefault("requests", []).append({"path": "/"})
        result = self.post(requests)

        self.assertEqual(
            self.request.errors[0]["description"], "Number of requests is limited to 22"
        )
        self.assertIsNone(result)  # rest of view not executed



import logging

import colander
from cornice.validators import colander_validator
from pyramid import httpexceptions
from pyramid.security import NO_PERMISSION_REQUIRED

from kinto.core import Service, errors
from kinto.core.errors import ErrorSchema
from kinto.core.resource.viewset import CONTENT_TYPES
from kinto.core.utils import build_request, build_response, merge_dicts

subrequest_logger = logging.getLogger("subrequest.summary")

valid_http_method = colander.OneOf(("GET", "HEAD", "DELETE", "TRACE", "POST", "PUT", "PATCH"))


def string_values(node, cstruct):
    """Validate that a ``colander.Mapping`` only has strings in its values.

    .. warning::

        Should be associated to a ``colander.Mapping`` schema node.
    """
    are_strings = [isinstance(v, str) for v in cstruct.values()]
    if not all(are_strings):
        error_msg = f"{cstruct} contains non string value"
        raise colander.Invalid(node, error_msg)


class BatchRequestSchema(colander.MappingSchema):
    method = colander.SchemaNode(
        colander.String(), validator=valid_http_method, missing=colander.drop
    )
    path = colander.SchemaNode(colander.String(), validator=colander.Regex("^/"))
    headers = colander.SchemaNode(
        colander.Mapping(unknown="preserve"), validator=string_values, missing=colander.drop
    )
    body = colander.SchemaNode(colander.Mapping(unknown="preserve"), missing=colander.drop)

    @staticmethod
    def schema_type():
        return colander.Mapping(unknown="raise")


class BatchPayloadSchema(colander.MappingSchema):
    defaults = BatchRequestSchema(missing=colander.drop).clone()
    requests = colander.SchemaNode(colander.Sequence(), BatchRequestSchema())

    @staticmethod
    def schema_type():
        return colander.Mapping(unknown="raise")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # On defaults, path is not mandatory.
        self.get("defaults").get("path").missing = colander.drop

    def deserialize(self, cstruct=colander.null):
        """Preprocess received data to carefully merge defaults.
        """
        if cstruct is not colander.null:
            defaults = cstruct.get("defaults")
            requests = cstruct.get("requests")
            if isinstance(defaults, dict) and isinstance(requests, list):
                for request in requests:
                    if isinstance(request, dict):
                        merge_dicts(request, defaults)
        return super().deserialize(cstruct)


class BatchRequest(colander.MappingSchema):
    body = BatchPayloadSchema()


class BatchResponseSchema(colander.MappingSchema):
    status = colander.SchemaNode(colander.Integer())
    path = colander.SchemaNode(colander.String())
    headers = colander.SchemaNode(
        colander.Mapping(unknown="preserve"), validator=string_values, missing=colander.drop
    )
    body = colander.SchemaNode(colander.Mapping(unknown="preserve"), missing=colander.drop)


class BatchResponseBodySchema(colander.MappingSchema):
    responses = colander.SequenceSchema(BatchResponseSchema(missing=colander.drop))


class BatchResponse(colander.MappingSchema):
    body = BatchResponseBodySchema()


class ErrorResponseSchema(colander.MappingSchema):
    body = ErrorSchema()


batch_responses = {
    "200": BatchResponse(description="Return a list of operation responses."),
    "400": ErrorResponseSchema(description="The request was badly formatted."),
    "default": ErrorResponseSchema(description="an unknown error occurred."),
}

batch = Service(name="batch", path="/batch", description="Batch operations")


@batch.post(
    schema=BatchRequest(),
    validators=(colander_validator,),
    content_type=CONTENT_TYPES,
    permission=NO_PERMISSION_REQUIRED,
    tags=["Batch"],
    operation_id="batch",
    response_schemas=batch_responses,
)
def post_batch(request):
    requests = request.validated["body"]["requests"]

    request.log_context(batch_size=len(requests))

    limit = request.registry.settings["batch_max_requests"]
    if limit and len(requests) > int(limit):
        error_msg = f"Number of requests is limited to {limit}"
        request.errors.add("body", "requests", error_msg)
        return

    if any([batch.path in req["path"] for req in requests]):
        error_msg = f"Recursive call on {batch.path} endpoint is forbidden."
        request.errors.add("body", "requests", error_msg)
        return

    responses = []

    for subrequest_spec in requests:
        subrequest = build_request(request, subrequest_spec)

        log_context = {
            **request.log_context(),
            "path": subrequest.path,
            "method": subrequest.method,
        }
        try:
            # Invoke subrequest without individual transaction.
            resp, subrequest = request.follow_subrequest(subrequest, use_tweens=False)
        except httpexceptions.HTTPException as e:
            # Since some request in the batch failed, we need to stop the parent request
            # through Pyramid's transaction manager. 5XX errors are already caught by
            # pyramid_tm's commit_veto
            # https://github.com/Kinto/kinto/issues/624
            if e.status_code == 409:
                request.tm.abort()

            if e.content_type == "application/json":
                resp = e
            else:
                # JSONify raw Pyramid errors.
                resp = errors.http_error(e)

        subrequest_logger.info("subrequest.summary", extra=log_context)

        dict_resp = build_response(resp, subrequest)
        responses.append(dict_resp)

    return {"responses": responses}



from bravado_core.response import validate_response

from .support import MINIMALIST_BUCKET, OpenAPITest


class OpenAPIBucketResponsesTest(OpenAPITest):
    def test_get_bucket_200(self):
        response = self.app.get("/buckets/b1", headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].get_bucket
        schema = self.spec.deref(op.op_spec["responses"]["200"])
        validate_response(schema, op, response)

    def test_post_bucket_200(self):
        response = self.app.post_json("/buckets", self.bucket, headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].create_bucket
        schema = self.spec.deref(op.op_spec["responses"]["200"])
        validate_response(schema, op, response)

    def test_post_bucket_201(self):
        response = self.app.post_json(
            "/buckets", MINIMALIST_BUCKET, headers=self.headers, status=201
        )
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].create_bucket
        schema = self.spec.deref(op.op_spec["responses"]["201"])
        validate_response(schema, op, response)

    def test_put_bucket_200(self):
        response = self.app.put("/buckets/b1", headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].update_bucket
        schema = self.spec.deref(op.op_spec["responses"]["200"])
        validate_response(schema, op, response)

    def test_put_bucket_201(self):
        response = self.app.put("/buckets/b2", headers=self.headers, status=201)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].update_bucket
        schema = self.spec.deref(op.op_spec["responses"]["201"])
        validate_response(schema, op, response)

    def test_delete_bucket_200(self):
        response = self.app.delete("/buckets/b1", headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].delete_bucket
        schema = self.spec.deref(op.op_spec["responses"]["200"])
        validate_response(schema, op, response)

    def test_get_buckets_200(self):
        response = self.app.get("/buckets", headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].get_buckets
        schema = self.spec.deref(op.op_spec["responses"]["200"])
        validate_response(schema, op, response)

    def test_delete_buckets_200(self):
        response = self.app.delete("/buckets", headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].delete_buckets
        schema = self.spec.deref(op.op_spec["responses"]["200"])
        validate_response(schema, op, response)



import unittest
from unittest import mock

from kinto.core.permission import PermissionBase
from kinto.core.permission import memory as memory_backend
from kinto.core.permission import postgresql as postgresql_backend
from kinto.core.permission.testing import PermissionTest
from kinto.core.testing import skip_if_no_postgresql
from kinto.core.utils import sqlalchemy


class PermissionBaseTest(unittest.TestCase):
    def setUp(self):
        self.permission = PermissionBase()

    def test_mandatory_overrides(self):
        calls = [
            (self.permission.initialize_schema,),
            (self.permission.flush,),
            (self.permission.add_user_principal, "", ""),
            (self.permission.remove_user_principal, "", ""),
            (self.permission.remove_principal, ""),
            (self.permission.get_user_principals, ""),
            (self.permission.add_principal_to_ace, "", "", ""),
            (self.permission.remove_principal_from_ace, "", "", ""),
            (self.permission.get_object_permission_principals, "", ""),
            (self.permission.get_objects_permissions, ""),
            (self.permission.replace_object_permissions, "", {}),
            (self.permission.delete_object_permissions, ""),
            (self.permission.get_accessible_objects, [], ""),
            (self.permission.get_authorized_principals, []),
        ]
        for call in calls:
            self.assertRaises(NotImplementedError, *call)


class MemoryPermissionTest(PermissionTest, unittest.TestCase):
    backend = memory_backend

    def test_backend_error_is_raised_anywhere(self):
        pass

    def test_ping_returns_false_if_unavailable(self):
        pass

    def test_ping_logs_error_if_unavailable(self):
        pass


@skip_if_no_postgresql
class PostgreSQLPermissionTest(PermissionTest, unittest.TestCase):
    backend = postgresql_backend
    settings = {
        "permission_backend": "kinto.core.permission.postgresql",
        "permission_pool_size": 10,
        "permission_url": "postgresql://postgres:postgres@localhost:5432/testdb",
    }

    def setUp(self):
        super().setUp()
        self.client_error_patcher = [
            mock.patch.object(
                self.permission.client,
                "session_factory",
                side_effect=sqlalchemy.exc.SQLAlchemyError,
            )
        ]



import getpass
import logging

import transaction as current_transaction
from pyramid.settings import asbool

from .utils import hash_password
from .views import AccountIdGenerator

logger = logging.getLogger(__name__)


def create_user(env, username=None, password=None):
    """Administrative command to create a new user."""
    registry = env["registry"]
    settings = registry.settings
    readonly_mode = asbool(settings.get("readonly", False))
    if readonly_mode:
        message = "Cannot create a user with a readonly server."
        logger.error(message)
        return 51

    if "kinto.plugins.accounts" not in settings["includes"]:
        message = "Cannot create a user when the accounts plugin is not installed."
        logger.error(message)
        return 52

    try:
        validator = AccountIdGenerator()
        if username is None:
            username = input("Username: ")
        while not validator.match(username):
            print("{} is not a valid username.")
            print(f"Username should match {validator.regexp}, please try again.")
            username = input("Username: ")

        if password is None:
            while True:  # The user didn't entered twice the same password
                password = getpass.getpass(f"Please enter a password for {username}: ")
                confirm = getpass.getpass("Please confirm the password: ")

                if password == confirm:
                    break
                print("Sorry, passwords do not match, please try again.")
    except EOFError:
        print("User creation aborted")
        return 53

    print(f"Creating user '{username}'")
    entry = {"id": username, "password": hash_password(password)}
    registry.storage.update(
        resource_name="account", parent_id=username, object_id=username, obj=entry
    )
    registry.permission.add_principal_to_ace(
        f"/accounts/{username}", "write", f"account:{username}"
    )

    current_transaction.commit()

    return 0



import bcrypt
from pyramid import authentication as base_auth

from kinto.core import utils
from kinto.core.storage import exceptions as storage_exceptions

from .utils import (
    ACCOUNT_POLICY_NAME,
    cache_account,
    delete_cached_reset_password,
    get_account_cache_key,
    get_cached_account,
    get_cached_reset_password,
    is_validated,
    refresh_cached_account,
)


def account_check(username, password, request):
    settings = request.registry.settings
    validation_enabled = settings.get("account_validation", False)
    cache_key = get_account_cache_key(username, request.registry)
    hashed_password = utils.hmac_digest(cache_key, password)

    # Check cache to see whether somebody has recently logged in with the same
    # username and password.
    cache_result = get_cached_account(username, request.registry)

    # Username and password have been verified previously. No need to compare hashes
    if cache_result == hashed_password:
        # Refresh the cache TTL.
        refresh_cached_account(username, request.registry)
        return True

    # Back to standard procedure
    parent_id = username
    try:
        existing = request.registry.storage.get(
            parent_id=parent_id, resource_name="account", object_id=username
        )
    except storage_exceptions.ObjectNotFoundError:
        return None

    if validation_enabled and not is_validated(existing):
        return None

    hashed = existing["password"].encode(encoding="utf-8")
    pwd_str = password.encode(encoding="utf-8")
    # Check if password is valid (it is a very expensive computation)
    if bcrypt.checkpw(pwd_str, hashed):
        cache_account(hashed_password, username, request.registry)
        return True

    # Last chance, is this a "reset password" flow?
    return reset_password_flow(username, password, request)


def reset_password_flow(username, password, request):
    cache_key = get_account_cache_key(username, request.registry)
    hashed_password = utils.hmac_digest(cache_key, password)
    pwd_str = password.encode(encoding="utf-8")

    cached_password = get_cached_reset_password(username, request.registry)
    if not cached_password:
        return None

    # The temporary reset password is only available for changing a user's password.
    if request.method.lower() not in ["post", "put", "patch"]:
        return None

    # Only allow modifying a user account, no other resource.
    uri = utils.strip_uri_prefix(request.path)
    resource_name, _ = utils.view_lookup(request, uri)
    if resource_name != "account":
        return None

    try:
        data = request.json["data"]
    except (ValueError, KeyError):
        return None

    # Request one and only one data field: the `password`.
    if not data or "password" not in data or len(data.keys()) > 1:
        return None

    cached_password_str = cached_password.encode(encoding="utf-8")
    if bcrypt.checkpw(pwd_str, cached_password_str):
        # Remove the temporary reset password from the cache.
        delete_cached_reset_password(username, request.registry)
        cache_account(hashed_password, username, request.registry)
        return True


class AccountsAuthenticationPolicy(base_auth.BasicAuthAuthenticationPolicy):
    """Accounts authentication policy.

    It will check that the credentials exist in the account resource.
    """

    name = ACCOUNT_POLICY_NAME

    def __init__(self, *args, **kwargs):
        super().__init__(account_check, *args, **kwargs)

    def effective_principals(self, request):
        # Bypass default Pyramid construction of principals because
        # Pyramid multiauth already adds userid, Authenticated and Everyone
        # principals.
        return []



from kinto.core import Service
from kinto.core.authorization import RouteFactory

attachment = Service(
    name="attachment",
    description="Attach file to object",
    path="/attachment",
    factory=RouteFactory,
)


@attachment.post(permission="attach")
def attachment_post(request):
    return {"ok": True}


def includeme(config):
    config.add_cornice_service(attachment)



"""
This module is a release hook for `zest.releaser <https://zestreleaser.readthedocs.io>`_
in order to build the Kinto Admin UI bundle just before it is packaged.

* See hooks in :file:`setup.cfg`
* `Documentation <http://zestreleaser.readthedocs.io/en/latest/entrypoints.html>`_
"""

import subprocess


def after_checkout(data):
    """
    During the ``release`` process, the current tag is checked out in a
    temporary folder. We build the Kinto Admin at this step, just before
    the files are gathered for the final Python package.

    .. note::

        The ``node_modules`` folder is excluded using :file:`MANIFEST.in`.
    """
    subprocess.run(["make", "build-kinto-admin"])



import functools
import logging
import re
import warnings
from uuid import uuid4

import colander
import venusian
from pyramid import exceptions as pyramid_exceptions
from pyramid.decorator import reify
from pyramid.httpexceptions import (
    HTTPNotFound,
    HTTPNotModified,
    HTTPPreconditionFailed,
    HTTPServiceUnavailable,
)
from pyramid.security import Everyone

from kinto.core import Service
from kinto.core.errors import ERRORS, http_error, raise_invalid, request_GET, send_alert
from kinto.core.events import ACTIONS
from kinto.core.storage import MISSING, Filter, Sort
from kinto.core.storage import exceptions as storage_exceptions
from kinto.core.utils import (
    COMPARISON,
    apply_json_patch,
    classname,
    decode64,
    dict_subset,
    encode64,
    find_nested_value,
    json,
    recursive_update_dict,
)

from .model import Model
from .schema import JsonPatchRequestSchema, ResourceSchema
from .viewset import ViewSet

logger = logging.getLogger(__name__)


def register(depth=1, **kwargs):
    """Ressource class decorator.

    Register the decorated class in the cornice registry.
    Pass all its keyword arguments to the register_resource
    function.
    """

    def wrapped(resource):
        register_resource(resource, depth=depth + 1, **kwargs)
        return resource

    return wrapped


def register_resource(resource_cls, settings=None, viewset=None, depth=1, **kwargs):
    """Register a resource in the cornice registry.

    :param resource_cls:
        The resource class to register.
        It should be a class or have a "name" attribute.

    :param viewset:
        A ViewSet object, which will be used to find out which arguments should
        be appended to the views, and where the views are.

    :param depth:
        A depth offset. It will be used to determine what is the level of depth
        in the call tree. (set to 1 by default.)

    Any additional keyword parameters will be used to override the viewset
    attributes.
    """
    if viewset is None:
        viewset = resource_cls.default_viewset(**kwargs)
    else:
        viewset.update(**kwargs)

    resource_name = viewset.get_name(resource_cls)

    def register_service(endpoint_type, settings):
        """Registers a service in cornice, for the given type.
        """
        path_pattern = getattr(viewset, f"{endpoint_type}_path")
        path_values = {"resource_name": resource_name}
        path = path_pattern.format_map(path_values)

        name = viewset.get_service_name(endpoint_type, resource_cls)

        service = Service(name, path, depth=depth, **viewset.get_service_arguments())

        # Attach viewset and resource to the service for later reference.
        service.viewset = viewset
        service.resource = resource_cls
        service.type = endpoint_type
        # Attach plural and object paths.
        service.plural_path = viewset.plural_path.format_map(path_values)
        service.object_path = (
            viewset.object_path.format_map(path_values)
            if viewset.object_path is not None
            else None
        )

        methods = getattr(viewset, f"{endpoint_type}_methods")
        for method in methods:
            if not viewset.is_endpoint_enabled(
                endpoint_type, resource_name, method.lower(), settings
            ):
                continue

            argument_getter = getattr(viewset, f"{endpoint_type}_arguments")
            view_args = argument_getter(resource_cls, method)

            view = viewset.get_view(endpoint_type, method.lower())
            service.add_view(method, view, klass=resource_cls, **view_args)

            # We support JSON-patch on PATCH views. Since the body payload
            # of JSON Patch is not a dict (mapping) but an array, we can't
            # use the same schema as for other PATCH protocols. We add another
            # dedicated view for PATCH, but targetting a different content_type
            # predicate.
            if method.lower() == "patch":
                view_args["content_type"] = "application/json-patch+json"
                view_args["schema"] = JsonPatchRequestSchema()
                service.add_view(method, view, klass=resource_cls, **view_args)

        return service

    def callback(context, name, ob):
        # get the callbacks registred by the inner services
        # and call them from here when the @resource classes are being
        # scanned by venusian.
        config = context.config.with_package(info.module)

        # Storage is mandatory for resources.
        if not hasattr(config.registry, "storage"):
            msg = "Mandatory storage backend is missing from configuration."
            raise pyramid_exceptions.ConfigurationError(msg)

        # A service for the list.
        service = register_service("plural", config.registry.settings)
        config.add_cornice_service(service)
        # An optional one for object endpoint.
        if getattr(viewset, "object_path") is not None:
            service = register_service("object", config.registry.settings)
            config.add_cornice_service(service)

    info = venusian.attach(resource_cls, callback, category="pyramid", depth=depth)
    return callback


class Resource:
    """Resource class providing every HTTP endpoint.

    A resource provides all the necessary mechanism for:
    - storage and retrieval of objects according to HTTP verbs
    - permission checking and tracking
    - concurrency control
    - synchronization
    - OpenAPI metadata

    Permissions are verified in :class:`kinto.core.authorization.AuthorizationPolicy` based on the
    verb and context (eg. a put can create or update). The resulting context
    is passed in the `context` constructor parameter.
    """

    default_viewset = ViewSet
    """Default :class:`kinto.core.resource.viewset.ViewSet` class to use when
    the resource is registered."""

    default_model = Model
    """Default :class:`kinto.core.resource.model.Model` class to use for
    interacting the :mod:`kinto.core.storage` and :mod:`kinto.core.permission`
    backends."""

    schema = ResourceSchema
    """Schema to validate objects."""

    permissions = ("read", "write")
    """List of allowed permissions names."""

    def __init__(self, request, context=None):
        """
        :param request:
            The current request object.
        :param context:
            The resulting context obtained from :class:`kinto.core.authorization.AuthorizationPolicy`.
        """
        self.request = request
        self.context = context

        content_type = str(self.request.headers.get("Content-Type")).lower()
        self._is_json_patch = content_type == "application/json-patch+json"
        self._is_merge_patch = content_type == "application/merge-patch+json"

        # Models are isolated by user.
        parent_id = self.get_parent_id(request)

        # The principal of an anonymous is system.Everyone
        current_principal = self.request.prefixed_userid or Everyone

        if not hasattr(self, "model"):
            self.model = self.default_model(
                storage=request.registry.storage,
                permission=request.registry.permission,
                id_generator=self.id_generator,
                resource_name=classname(self),
                parent_id=parent_id,
                current_principal=current_principal,
                prefixed_principals=request.prefixed_principals,
            )

        # Initialize timestamp as soon as possible.
        self.timestamp

        if self.context:
            self.model.get_permission_object_id = functools.partial(
                self.context.get_permission_object_id, self.request
            )

    @reify
    def id_generator(self):
        # ID generator by resource name in settings.
        default_id_generator = self.request.registry.id_generators[""]
        resource_name = self.request.current_resource_name
        id_generator = self.request.registry.id_generators.get(resource_name, default_id_generator)
        return id_generator

    @reify
    def timestamp(self):
        """Return the current resource timestamp.

        :rtype: int
        """
        try:
            return self.model.timestamp()
        except storage_exceptions.BackendError as e:
            is_readonly = self.request.registry.settings["readonly"]
            if not is_readonly:
                raise e
            # If the instance is configured to be readonly, and if the
            # resource is empty, the backend will try to bump the timestamp.
            # It fails if the configured db user has not write privileges.
            logger.exception(e)
            error_msg = (
                "Resource timestamp cannot be written. "
                "Plural endpoint must be hit at least once from a "
                "writable instance."
            )
            raise http_error(HTTPServiceUnavailable(), errno=ERRORS.BACKEND, message=error_msg)

    @reify
    def object_id(self):
        """Return the object id for this request. It's either in the match dict
        or in the posted body.
        """
        if self.request.method.lower() == "post":
            try:
                # Since ``id`` does not belong to schema, it is not in validated
                # data. Must look up in body directly instead of request.validated.
                _id = self.request.json["data"][self.model.id_field]
                self._raise_400_if_invalid_id(_id)
                return _id
            except (KeyError, ValueError):
                return None
        return self.request.matchdict.get("id")

    def get_parent_id(self, request):
        """Return the parent_id of the resource with regards to the current
        request.

        The resource will isolate the objects from one parent id to another.
        For example, in Kinto, the ``group``s and ``collection``s are isolated by ``bucket``.

        In order to obtain a resource where users can only see their own objects, just
        return the user id as the parent id:

        .. code-block:: python

            def get_parent_id(self, request):
                return request.prefixed_userid

        :param request:
            The request used to access the resource.

        :rtype: str
        """
        return ""

    def _get_known_fields(self):
        """Return all the `field` defined in the ressource schema."""
        known_fields = [c.name for c in self.schema().children] + [
            self.model.id_field,
            self.model.modified_field,
            self.model.deleted_field,
        ]
        return known_fields

    def is_known_field(self, field):
        """Return ``True`` if `field` is defined in the resource schema.
        If the resource schema allows unknown fields, this will always return
        ``True``.

        :param str field: Field name
        :rtype: bool

        """
        if self.schema.get_option("preserve_unknown"):
            return True

        known_fields = self._get_known_fields()
        # Test first level only: ``target.data.id`` -> ``target``
        field = field.split(".", 1)[0]
        return field in known_fields

    #
    # End-points
    #

    def plural_head(self):
        """Model ``HEAD`` endpoint: empty reponse with a ``Total-Objects`` header.

        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPNotModified` if
            ``If-None-Match`` header is provided and collection not
            modified in the interim.

        :raises:
            :exc:`~pyramid:pyramid.httpexceptions.HTTPPreconditionFailed` if
            ``If-Match`` header is provided and collection modified
            in the iterim.
        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPBadRequest`
            if filters or sorting are invalid.
        """
        return self._plural_get(True)

    def plural_get(self):
        """Model ``GET`` endpoint: retrieve multiple objects.

        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPNotModified` if
            ``If-None-Match`` header is provided and the objects not
            modified in the interim.

        :raises:
            :exc:`~pyramid:pyramid.httpexceptions.HTTPPreconditionFailed` if
            ``If-Match`` header is provided and the objects modified
            in the iterim.
        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPBadRequest`
            if filters or sorting are invalid.
        """
        return self._plural_get(False)

    def _plural_get(self, head_request):
        self._add_timestamp_header(self.request.response)
        self._add_cache_header(self.request.response)
        self._raise_304_if_not_modified()
        # Plural endpoints are considered resources that always exist
        self._raise_412_if_modified(obj={})

        headers = self.request.response.headers

        filters = self._extract_filters()
        limit = self._extract_limit()
        sorting = self._extract_sorting(limit)
        partial_fields = self._extract_partial_fields()

        filter_fields = [f.field for f in filters]
        include_deleted = self.model.modified_field in filter_fields

        pagination_rules, offset = self._extract_pagination_rules_from_token(limit, sorting)

        # The reason why we call self.model.get_objects() with `limit=limit + 1` is to avoid
        # having to count the total number of objects in the database just to be able
        # to *decide* whether or not to have a `Next-Page` header.
        # This way, we can quickly depend on the number of objects returned and compare that
        # with what the client requested.
        # For example, if there are 100 objects in the database and the client used limit=100,
        # it would, internally, ask for 101 objects. So if you retrieved 100 objects
        # it means we got less than we asked for and thus there is not another page.
        # Equally, if there are 200 objects in the database and the client used
        # limit=100 it would, internally, ask for 101 objects and actually get that. Then,
        # you know there is another page.

        if head_request:
            count = self.model.count_objects(filters=filters)
            headers["Total-Objects"] = headers["Total-Records"] = str(count)
            return self.postprocess([])

        objects = self.model.get_objects(
            filters=filters,
            sorting=sorting,
            limit=limit + 1,  # See bigger explanation above.
            pagination_rules=pagination_rules,
            include_deleted=include_deleted,
        )

        offset = offset + len(objects)

        if limit and len(objects) == limit + 1:
            lastobject = objects[-2]
            next_page = self._next_page_url(sorting, limit, lastobject, offset)
            headers["Next-Page"] = next_page

        if partial_fields:
            objects = [dict_subset(obj, partial_fields) for obj in objects]

        # See bigger explanation above about the use of limits. The need for slicing
        # here is because we might have asked for 1 more object just to see if there's
        # a next page. But we have to honor the limit in our returned response.
        return self.postprocess(objects[:limit])

    def plural_post(self):
        """Model ``POST`` endpoint: create an object.

        If the new object id conflicts against an existing one, the
        posted object is ignored, and the existing object is returned, with
        a ``200`` status.

        :raises:
            :exc:`~pyramid:pyramid.httpexceptions.HTTPPreconditionFailed` if
            ``If-Match`` header is provided and the objects modified
            in the iterim.

        .. seealso::

            Add custom behaviour by overriding
            :meth:`kinto.core.resource.Resource.process_object`
        """
        new_object = self.request.validated["body"].get("data", {})

        existing = None
        # If id was specified, then add it to posted body and look-up
        # the existing object.
        if self.object_id is not None:
            new_object[self.model.id_field] = self.object_id
            try:
                existing = self._get_object_or_404(self.object_id)
            except HTTPNotFound:
                pass

        self._raise_412_if_modified(obj=existing)

        if existing:
            obj = existing
            action = ACTIONS.READ
        else:
            new_object = self.process_object(new_object)
            obj = self.model.create_object(new_object)
            self.request.response.status_code = 201
            action = ACTIONS.CREATE

        timestamp = obj[self.model.modified_field]
        self._add_timestamp_header(self.request.response, timestamp=timestamp)

        return self.postprocess(obj, action=action)

    def plural_delete(self):
        """Model ``DELETE`` endpoint: delete multiple objects.

        :raises:
            :exc:`~pyramid:pyramid.httpexceptions.HTTPPreconditionFailed` if
            ``If-Match`` header is provided and the objects modified
            in the iterim.

        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPBadRequest`
            if filters are invalid.
        """
        # Plural endpoint are considered resources that always exist
        self._raise_412_if_modified(obj={})

        filters = self._extract_filters()
        limit = self._extract_limit()
        sorting = self._extract_sorting(limit)
        pagination_rules, offset = self._extract_pagination_rules_from_token(limit, sorting)

        objects = self.model.get_objects(
            filters=filters, sorting=sorting, limit=limit + 1, pagination_rules=pagination_rules
        )
        deleted = self.model.delete_objects(
            filters=filters, sorting=sorting, limit=limit, pagination_rules=pagination_rules
        )
        if deleted:
            lastobject = deleted[-1]
            # Add pagination header, but only if there are more objects beyond the limit.
            if limit and len(objects) == limit + 1:
                next_page = self._next_page_url(sorting, limit, lastobject, offset)
                self.request.response.headers["Next-Page"] = next_page

            timestamp = max({d[self.model.modified_field] for d in deleted})
            self._add_timestamp_header(self.request.response, timestamp=timestamp)

        else:
            self._add_timestamp_header(self.request.response)

        action = len(deleted) > 0 and ACTIONS.DELETE or ACTIONS.READ
        return self.postprocess(deleted, action=action, old=objects[:limit])

    def get(self):
        """Object ``GET`` endpoint: retrieve an object.

        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPNotFound` if
            the object is not found.

        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPNotModified` if
            ``If-None-Match`` header is provided and object not
            modified in the interim.

        :raises:
            :exc:`~pyramid:pyramid.httpexceptions.HTTPPreconditionFailed` if
            ``If-Match`` header is provided and object modified
            in the iterim.
        """
        self._raise_400_if_invalid_id(self.object_id)
        obj = self._get_object_or_404(self.object_id)
        timestamp = obj[self.model.modified_field]
        self._add_timestamp_header(self.request.response, timestamp=timestamp)
        self._add_cache_header(self.request.response)
        self._raise_304_if_not_modified(obj)
        self._raise_412_if_modified(obj)

        partial_fields = self._extract_partial_fields()
        if partial_fields:
            obj = dict_subset(obj, partial_fields)

        return self.postprocess(obj)

    def put(self):
        """Object ``PUT`` endpoint: create or replace the provided object and
        return it.

        :raises:
            :exc:`~pyramid:pyramid.httpexceptions.HTTPPreconditionFailed` if
            ``If-Match`` header is provided and object modified
            in the iterim.

        .. note::

            If ``If-None-Match: *`` request header is provided, the
            ``PUT`` will succeed only if no object exists with this id.

        .. seealso::

            Add custom behaviour by overriding
            :meth:`kinto.core.resource.Resource.process_object`.
        """
        self._raise_400_if_invalid_id(self.object_id)
        try:
            existing = self._get_object_or_404(self.object_id)
        except HTTPNotFound:
            existing = None

        self._raise_412_if_modified(obj=existing)

        # If `data` is not provided, use existing object (or empty if creation)
        post_object = self.request.validated["body"].get("data", existing) or {}

        object_id = post_object.setdefault(self.model.id_field, self.object_id)
        self._raise_400_if_id_mismatch(object_id, self.object_id)

        new_object = self.process_object(post_object, old=existing)

        if existing:
            obj = self.model.update_object(new_object)
        else:
            obj = self.model.create_object(new_object)
            self.request.response.status_code = 201

        timestamp = obj[self.model.modified_field]
        self._add_timestamp_header(self.request.response, timestamp=timestamp)

        action = existing and ACTIONS.UPDATE or ACTIONS.CREATE
        return self.postprocess(obj, action=action, old=existing)

    def patch(self):
        """Object ``PATCH`` endpoint: modify an object and return its
        new version.

        If a request header ``Response-Behavior`` is set to ``light``,
        only the fields whose value was changed are returned.
        If set to ``diff``, only the fields whose value became different than
        the one provided are returned.

        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPNotFound` if
            the object is not found.

        :raises:
            :exc:`~pyramid:pyramid.httpexceptions.HTTPPreconditionFailed` if
            ``If-Match`` header is provided and object modified
            in the iterim.

        .. seealso::
            Add custom behaviour by overriding
            :meth:`kinto.core.resource.Resource.apply_changes` or
            :meth:`kinto.core.resource.Resource.process_object`.
        """
        self._raise_400_if_invalid_id(self.object_id)
        existing = self._get_object_or_404(self.object_id)
        self._raise_412_if_modified(existing)

        # patch is specified as a list of of operations (RFC 6902)
        if self._is_json_patch:
            requested_changes = self.request.validated["body"]
        else:
            # `data` attribute may not be present if only perms are patched.
            body = self.request.validated["body"]
            if not body:
                # If no `data` nor `permissions` is provided in patch, reject!
                # XXX: This should happen in schema instead (c.f. ViewSet)
                error_details = {
                    "name": "data",
                    "description": "Provide at least one of data or permissions",
                }
                raise_invalid(self.request, **error_details)
            requested_changes = body.get("data", {})

        updated, applied_changes = self.apply_changes(
            obj=existing, requested_changes=requested_changes
        )

        object_id = updated.setdefault(self.model.id_field, self.object_id)
        self._raise_400_if_id_mismatch(object_id, self.object_id)

        new_object = self.process_object(updated, old=existing)

        changed_fields = [
            k for k in applied_changes.keys() if existing.get(k) != new_object.get(k)
        ]

        new_object = self.model.update_object(new_object)

        # Adjust response according to ``Response-Behavior`` header
        body_behavior = self.request.validated["header"].get("Response-Behavior", "full")

        if body_behavior.lower() == "light":
            # Only fields that were changed.
            data = {k: new_object[k] for k in changed_fields}

        elif body_behavior.lower() == "diff":
            # Only fields that are different from those provided.
            data = {
                k: new_object[k]
                for k in changed_fields
                if applied_changes.get(k) != new_object.get(k)
            }
        else:
            data = new_object

        timestamp = new_object.get(self.model.modified_field, existing[self.model.modified_field])
        self._add_timestamp_header(self.request.response, timestamp=timestamp)

        return self.postprocess(data, action=ACTIONS.UPDATE, old=existing)

    def delete(self):
        """Object ``DELETE`` endpoint: delete an object and return it.

        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPNotFound` if
            the object is not found.

        :raises:
            :exc:`~pyramid:pyramid.httpexceptions.HTTPPreconditionFailed` if
            ``If-Match`` header is provided and object modified
            in the iterim.
        """
        self._raise_400_if_invalid_id(self.object_id)
        obj = self._get_object_or_404(self.object_id)
        self._raise_412_if_modified(obj)

        # Retreive the last_modified information from a querystring if present.
        last_modified = self.request.validated["querystring"].get("last_modified")

        # If less or equal than current object. Ignore it.
        if last_modified and last_modified <= obj[self.model.modified_field]:
            last_modified = None

        try:
            deleted = self.model.delete_object(obj, last_modified=last_modified)
        except storage_exceptions.ObjectNotFoundError:
            # Delete might fail if the object was deleted since we
            # fetched it from the storage (ref Kinto/kinto#1407). This
            # is one of a larger class of issues where another request
            # could modify the object between our fetch and our
            # delete, which could e.g. invalidate our precondition
            # checking. Fixing this correctly is a larger
            # problem. However, let's punt on fixing it correctly and
            # just handle this one important case for now (see #1557).
            #
            # Raise a 404 vs. a 409 or 412 because that's what we
            # would have done if the other thread's delete had
            # happened a little earlier. (The client doesn't need to
            # know that we did a bunch of work fetching the existing
            # object for nothing.)
            raise self._404_for_object(self.object_id)

        timestamp = deleted[self.model.modified_field]
        self._add_timestamp_header(self.request.response, timestamp=timestamp)

        return self.postprocess(deleted, action=ACTIONS.DELETE, old=obj)

    #
    # Data processing
    #

    def process_object(self, new, old=None):
        """Hook for processing objects before they reach storage, to introduce
        specific logics on fields for example.

        .. code-block:: python

            def process_object(self, new, old=None):
                new = super().process_object(new, old)
                version = old['version'] if old else 0
                new['version'] = version + 1
                return new

        Or add extra validation based on request:

        .. code-block:: python

            from kinto.core.errors import raise_invalid

            def process_object(self, new, old=None):
                new = super().process_object(new, old)
                if new['browser'] not in request.headers['User-Agent']:
                    raise_invalid(self.request, name='browser', error='Wrong')
                return new

        :param dict new: the validated object to be created or updated.
        :param dict old: the old object to be updated,
            ``None`` for creation endpoints.

        :returns: the processed object.
        :rtype: dict
        """
        modified_field = self.model.modified_field
        new_last_modified = new.get(modified_field)

        # Drop the new last_modified if it is not an integer.
        is_integer = isinstance(new_last_modified, int)
        if not is_integer:
            new.pop(modified_field, None)
            new_last_modified = None

        # Drop the new last_modified if lesser or equal to the old one.
        is_less_or_equal = (
            new_last_modified and old is not None and new_last_modified <= old[modified_field]
        )
        if is_less_or_equal:
            new.pop(modified_field, None)

        # patch is specified as a list of of operations (RFC 6902)

        payload = self.request.validated["body"]

        if self._is_json_patch:
            permissions = apply_json_patch(old, payload)["permissions"]

        elif self._is_merge_patch:
            existing = old or {}
            permissions = existing.get("__permissions__", {})
            recursive_update_dict(permissions, payload.get("permissions", {}), ignores=(None,))

        else:
            permissions = {
                k: v for k, v in payload.get("permissions", {}).items() if v is not None
            }

        annotated = {**new}

        if permissions:
            is_put = self.request.method.lower() == "put"
            if is_put or self._is_merge_patch:
                # Remove every existing ACEs using empty lists.
                for perm in self.permissions:
                    permissions.setdefault(perm, [])
            annotated[self.model.permissions_field] = permissions

        return annotated

    def apply_changes(self, obj, requested_changes):
        """Merge `changes` into `object` fields.

        .. note::

            This is used in the context of PATCH only.

        Override this to control field changes at object level, for example:

        .. code-block:: python

            def apply_changes(self, obj, requested_changes):
                # Ignore value change if inferior
                if object['position'] > changes.get('position', -1):
                    changes.pop('position', None)
                return super().apply_changes(obj, requested_changes)

        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPBadRequest`
            if result does not comply with resource schema.

        :returns: the new object with `changes` applied.
        :rtype: tuple
        """
        if self._is_json_patch:
            try:
                applied_changes = apply_json_patch(obj, requested_changes)["data"]
                updated = {**applied_changes}
            except ValueError as e:
                error_details = {
                    "location": "body",
                    "description": f"JSON Patch operation failed: {e}",
                }
                raise_invalid(self.request, **error_details)

        else:
            applied_changes = {**requested_changes}
            updated = {**obj}

            # recursive patch and remove field if null attribute is passed (RFC 7396)
            if self._is_merge_patch:
                recursive_update_dict(updated, applied_changes, ignores=(None,))
            else:
                updated.update(**applied_changes)

        for field, value in applied_changes.items():
            has_changed = obj.get(field, value) != value
            if self.schema.is_readonly(field) and has_changed:
                error_details = {"name": field, "description": f"Cannot modify {field}"}
                raise_invalid(self.request, **error_details)

        try:
            validated = self.schema().deserialize(updated)
        except colander.Invalid as e:
            # Transform the errors we got from colander into Cornice errors.
            # We could not rely on Service schema because the object should be
            # validated only once the changes are applied
            for field, error in e.asdict().items():  # pragma: no branch
                raise_invalid(self.request, name=field, description=error)

        return validated, applied_changes

    def postprocess(self, result, action=ACTIONS.READ, old=None):
        body = {}

        if not isinstance(result, list):
            perms = result.pop(self.model.permissions_field, None)
            if perms is not None:
                body["permissions"] = {k: list(p) for k, p in perms.items()}
            if old:
                # Remove permissions from event payload.
                old.pop(self.model.permissions_field, None)

        body["data"] = result

        parent_id = self.get_parent_id(self.request)
        # Use self.model.timestamp() instead of self.timestamp because
        # self.timestamp is @reify'd relatively early in the request,
        # so doesn't correspond to any time that is relevant to the
        # event. See #1769.
        timestamp = self.model.timestamp()
        self.request.notify_resource_event(
            parent_id=parent_id, timestamp=timestamp, data=result, action=action, old=old
        )

        return body

    #
    # Internals
    #

    def _404_for_object(self, object_id):
        details = {"id": object_id, "resource_name": self.request.current_resource_name}
        return http_error(HTTPNotFound(), errno=ERRORS.INVALID_RESOURCE_ID, details=details)

    def _get_object_or_404(self, object_id):
        """Retrieve object from storage and raise ``404 Not found`` if missing.

        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPNotFound` if
            the object is not found.
        """
        if self.context and self.context.current_object:
            # Set during authorization. Save a storage hit.
            return self.context.current_object

        try:
            return self.model.get_object(object_id)
        except storage_exceptions.ObjectNotFoundError:
            raise self._404_for_object(object_id)

    def _add_timestamp_header(self, response, timestamp=None):
        """Add current timestamp in response headers, when request comes in.

        """
        if timestamp is None:
            timestamp = self.timestamp
        # Pyramid takes care of converting.
        response.last_modified = timestamp / 1000.0
        # Return timestamp as ETag.
        response.headers["ETag"] = f'"{timestamp}"'

    def _add_cache_header(self, response):
        """Add Cache-Control and Expire headers, based a on a setting for the
        current resource.

        Cache headers will be set with anonymous requests only.

        .. note::

            The ``Cache-Control: no-cache`` response header does not prevent
            caching in client. It will indicate the client to revalidate
            the response content on each access. The client will send a
            conditional request to the server and check that a
            ``304 Not modified`` is returned before serving content from cache.
        """
        resource_name = self.context.resource_name if self.context else ""
        setting_key = f"{resource_name}_cache_expires_seconds"
        cache_expires = self.request.registry.settings.get(setting_key)
        is_anonymous = self.request.prefixed_userid is None
        if cache_expires and is_anonymous:
            response.cache_expires(seconds=int(cache_expires))
        else:
            # Since `Expires` response header provides an HTTP data with a
            # resolution in seconds, do not use Pyramid `cache_expires()` in
            # order to omit it.
            response.cache_control.no_cache = True
            response.cache_control.no_store = True

    def _raise_400_if_invalid_id(self, object_id):
        """Raise 400 if specified object id does not match the format excepted
        by storage backends.

        :raises: :class:`pyramid.httpexceptions.HTTPBadRequest`
        """
        is_string = isinstance(object_id, str)
        if not is_string or not self.model.id_generator.match(object_id):
            error_details = {"location": "path", "description": "Invalid object id"}
            raise_invalid(self.request, **error_details)

    def _raise_304_if_not_modified(self, obj=None):
        """Raise 304 if current timestamp is inferior to the one specified
        in headers.

        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPNotModified`
        """
        if_none_match = self.request.validated["header"].get("If-None-Match")

        if not if_none_match:
            return

        if if_none_match == "*":
            return

        if obj:
            current_timestamp = obj[self.model.modified_field]
        else:
            current_timestamp = self.model.timestamp()

        if current_timestamp == if_none_match:
            response = HTTPNotModified()
            self._add_timestamp_header(response, timestamp=current_timestamp)
            raise response

    def _raise_412_if_modified(self, obj=None):
        """Raise 412 if current timestamp is superior to the one
        specified in headers.

        :raises:
            :exc:`~pyramid:pyramid.httpexceptions.HTTPPreconditionFailed`
        """
        if_match = self.request.validated["header"].get("If-Match")
        if_none_match = self.request.validated["header"].get("If-None-Match")

        # Check if object exists
        object_exists = obj is not None

        # If no precondition headers, just ignore
        if not if_match and not if_none_match:
            return

        # If-None-Match: * should always raise if an object exists
        if if_none_match == "*" and object_exists:
            modified_since = -1  # Always raise.

        # If-Match should always raise if an object doesn't exist
        elif if_match and not object_exists:
            modified_since = -1

        # If-Match with ETag value on existing objects should compare ETag
        elif if_match and if_match != "*":
            modified_since = if_match

        # If none of the above applies, don't raise
        else:
            return

        if obj:
            current_timestamp = obj[self.model.modified_field]
        else:
            current_timestamp = self.model.timestamp()

        if current_timestamp != modified_since:
            error_msg = "Resource was modified meanwhile"
            # Do not provide the permissions among the object fields.
            # Ref: https://github.com/Kinto/kinto/issues/224
            existing = {**obj} if obj else {}
            existing.pop(self.model.permissions_field, None)

            details = {"existing": existing} if obj else {}
            response = http_error(
                HTTPPreconditionFailed(),
                errno=ERRORS.MODIFIED_MEANWHILE,
                message=error_msg,
                details=details,
            )
            self._add_timestamp_header(response, timestamp=current_timestamp)
            raise response

    def _raise_400_if_id_mismatch(self, new_id, object_id):
        """Raise 400 if the `new_id`, within the request body, does not match
        the `object_id`, obtained from request path.

        :raises: :class:`pyramid.httpexceptions.HTTPBadRequest`
        """
        if new_id != object_id:
            error_msg = "Object id does not match existing object"
            error_details = {"name": self.model.id_field, "description": error_msg}
            raise_invalid(self.request, **error_details)

    def _extract_partial_fields(self):
        """Extract the fields to do the projection from QueryString parameters.
        """
        fields = self.request.validated["querystring"].get("_fields")
        if fields:
            root_fields = [f.split(".")[0] for f in fields]
            known_fields = self._get_known_fields()
            invalid_fields = set(root_fields) - set(known_fields)
            preserve_unknown = self.schema.get_option("preserve_unknown")
            if not preserve_unknown and invalid_fields:
                error_msg = f"Fields {','.join(invalid_fields)} do not exist"
                error_details = {"name": "Invalid _fields parameter", "description": error_msg}
                raise_invalid(self.request, **error_details)

            # Since id and last_modified are part of the synchronisation
            # API, force their presence in payloads.
            fields = fields + [self.model.id_field, self.model.modified_field]

        return fields

    def _extract_limit(self):
        """Extract limit value from QueryString parameters."""
        paginate_by = self.request.registry.settings["paginate_by"]
        max_fetch_size = self.request.registry.settings["storage_max_fetch_size"]
        limit = self.request.validated["querystring"].get("_limit", paginate_by)

        # If limit is higher than paginate_by setting, ignore it.
        if limit and paginate_by:
            limit = min(limit, paginate_by)

        # If limit is higher than what storage can retrieve, ignore it.
        limit = min(limit, max_fetch_size) if limit else max_fetch_size

        return limit

    def _extract_filters(self):
        """Extracts filters from QueryString parameters."""
        queryparams = self.request.validated["querystring"]

        filters = []

        for param, value in queryparams.items():
            param = param.strip()

            error_details = {
                "name": param,
                "location": "querystring",
                "description": f"Invalid value for {param}",
            }

            # Ignore specific fields
            if param.startswith("_") and param not in ("_since", "_to", "_before"):
                continue

            # Handle the _since specific filter.
            if param in ("_since", "_to", "_before"):

                if param == "_since":
                    operator = COMPARISON.GT
                else:
                    if param == "_to":
                        message = "_to is now deprecated, " "you should use _before instead"
                        url = (
                            "https://kinto.readthedocs.io/en/2.4.0/api/"
                            "resource.html#list-of-available-url-"
                            "parameters"
                        )
                        send_alert(self.request, message, url)
                    operator = COMPARISON.LT

                if value == "":
                    raise_invalid(self.request, **error_details)

                filters.append(Filter(self.model.modified_field, value, operator))
                continue

            all_keywords = r"|".join([i.name.lower() for i in COMPARISON])
            m = re.match(r"^(" + all_keywords + r")_([\w\.]+)$", param)
            if m:
                keyword, field = m.groups()
                operator = getattr(COMPARISON, keyword.upper())
            else:
                operator, field = COMPARISON.EQ, param

            if not self.is_known_field(field):
                error_msg = f"Unknown filter field '{param}'"
                error_details["description"] = error_msg
                raise_invalid(self.request, **error_details)

            # Return 400 if _limit is not a string
            if operator == COMPARISON.LIKE:
                if not isinstance(value, str):
                    raise_invalid(self.request, **error_details)

            if operator in (COMPARISON.IN, COMPARISON.EXCLUDE):
                all_integers = all([isinstance(v, int) for v in value])
                all_strings = all([isinstance(v, str) for v in value])
                has_invalid_value = (field == self.model.id_field and not all_strings) or (
                    field == self.model.modified_field and not all_integers
                )
                if has_invalid_value:
                    raise_invalid(self.request, **error_details)

            if "\x00" in field or "\x00" in str(value):
                error_details["description"] = "Invalid character 0x00"
                raise_invalid(self.request, **error_details)

            if field == self.model.modified_field and value == "":
                raise_invalid(self.request, **error_details)

            filters.append(Filter(field, value, operator))

        # If a plural endpoint is reached, and if the user does not have the
        # permission to read/write the whole list, the set is filtered by ids,
        # based on the list of ids returned by the authorization policy.
        ids = self.context.shared_ids
        if ids is not None:
            filter_by_id = Filter(self.model.id_field, ids, COMPARISON.IN)
            filters.insert(0, filter_by_id)

        return filters

    def _extract_sorting(self, limit):
        """Extracts filters from QueryString parameters."""
        specified = self.request.validated["querystring"].get("_sort", [])
        sorting = []
        modified_field_used = self.model.modified_field in specified
        for field in specified:
            field = field.strip()
            m = re.match(r"^([\-+]?)([\w\.]+)$", field)
            if m:
                order, field = m.groups()

                if not self.is_known_field(field):
                    error_details = {
                        "location": "querystring",
                        "description": f"Unknown sort field '{field}'",
                    }
                    raise_invalid(self.request, **error_details)

                direction = -1 if order == "-" else 1
                sorting.append(Sort(field, direction))

        if not modified_field_used:
            # Add a sort by the ``modified_field`` in descending order
            # useful for pagination
            sorting.append(Sort(self.model.modified_field, -1))
        return sorting

    def _build_pagination_rules(self, sorting, last_object, rules=None):
        """Return the list of rules for a given sorting attribute and
        last_object.

        """
        if rules is None:
            rules = []

        rule = []
        next_sorting = sorting[:-1]

        for field, _ in next_sorting:
            rule.append(Filter(field, last_object.get(field, MISSING), COMPARISON.EQ))

        field, direction = sorting[-1]

        if direction == -1:
            rule.append(Filter(field, last_object.get(field, MISSING), COMPARISON.LT))
        else:
            rule.append(Filter(field, last_object.get(field, MISSING), COMPARISON.GT))

        rules.append(rule)

        if len(next_sorting) == 0:
            return rules

        return self._build_pagination_rules(next_sorting, last_object, rules)

    def _extract_pagination_rules_from_token(self, limit, sorting):
        """Get pagination params."""
        token = self.request.validated["querystring"].get("_token", None)
        filters = []
        offset = 0
        if token:
            error_msg = None
            try:
                tokeninfo = json.loads(decode64(token))
                if not isinstance(tokeninfo, dict):
                    raise ValueError()
                last_object = tokeninfo["last_object"]
                offset = tokeninfo["offset"]
                nonce = tokeninfo["nonce"]
            except (ValueError, KeyError, TypeError):
                error_msg = "_token has invalid content"

            # We don't want pagination tokens to be reused several times (#1171).
            # The cache backend is used to keep track of "nonces".
            if self.request.method.lower() == "delete" and error_msg is None:
                registry = self.request.registry
                deleted = registry.cache.delete(nonce)
                if deleted is None:
                    error_msg = "_token was already used or has expired."

            if error_msg:
                error_details = {"location": "querystring", "description": error_msg}
                raise_invalid(self.request, **error_details)

            filters = self._build_pagination_rules(sorting, last_object)

        return filters, offset

    def _next_page_url(self, sorting, limit, last_object, offset):
        """Build the Next-Page header from where we stopped."""
        token = self._build_pagination_token(sorting, last_object, offset)

        params = {**request_GET(self.request), "_limit": limit, "_token": token}

        service = self.request.current_service
        next_page_url = self.request.route_url(
            service.name, _query=params, **self.request.matchdict
        )
        return next_page_url

    def _build_pagination_token(self, sorting, last_object, offset):
        """Build a pagination token.

        It is a base64 JSON object with the sorting fields values of
        the last_object.

        """
        nonce = f"pagination-token-{uuid4()}"
        if self.request.method.lower() == "delete":
            registry = self.request.registry
            validity = registry.settings["pagination_token_validity_seconds"]
            registry.cache.set(nonce, "", validity)

        token = {"last_object": {}, "offset": offset, "nonce": nonce}

        for field, _ in sorting:
            last_value = find_nested_value(last_object, field, MISSING)
            if last_value is not MISSING:
                token["last_object"][field] = last_value

        return encode64(json.dumps(token))

    @property
    def record_id(self):
        message = "`record_id` is deprecated, use `object_id` instead."
        warnings.warn(message, DeprecationWarning)
        return self.object_id

    def process_record(self, *args, **kwargs):
        message = "`process_record()` is deprecated, use `process_object()` instead."
        warnings.warn(message, DeprecationWarning)
        return self.process_object(*args, **kwargs)

    def collection_get(self, *args, **kwargs):
        message = "`collection_get()` is deprecated, use `plural_get()` instead."
        warnings.warn(message, DeprecationWarning)
        return self.plural_get(*args, **kwargs)

    def collection_post(self, *args, **kwargs):
        message = "`collection_post()` is deprecated, use `plural_post()` instead."
        warnings.warn(message, DeprecationWarning)
        return self.plural_post(*args, **kwargs)

    def collection_delete(self, *args, **kwargs):
        message = "`collection_delete()` is deprecated, use `plural_delete()` instead."
        warnings.warn(message, DeprecationWarning)
        return self.plural_delete(*args, **kwargs)


class ShareableResource(Resource):
    def __init__(self, *args, **kwargs):
        message = "`ShareableResource` is deprecated, use `Resource` instead."
        warnings.warn(message, DeprecationWarning)
        super().__init__(*args, **kwargs)



import colander

from kinto.core import resource
from kinto.core.resource.viewset import ViewSet
from kinto.core.storage import Filter
from kinto.core.utils import instance_uri


class HistorySchema(resource.ResourceSchema):
    user_id = colander.SchemaNode(colander.String())
    uri = colander.SchemaNode(colander.String())
    action = colander.SchemaNode(colander.String())
    date = colander.SchemaNode(colander.String())
    resource_name = colander.SchemaNode(colander.String())
    bucket_id = colander.SchemaNode(colander.String(), missing=colander.drop)
    collection_id = colander.SchemaNode(colander.String(), missing=colander.drop)
    group_id = colander.SchemaNode(colander.String(), missing=colander.drop)
    record_id = colander.SchemaNode(colander.String(), missing=colander.drop)
    target = colander.SchemaNode(colander.Mapping())

    class Options:
        preserve_unknown = False


# Add custom OpenAPI tags/operation ids
plural_get_arguments = getattr(ViewSet, "plural_get_arguments", {})
plural_delete_arguments = getattr(ViewSet, "plural_delete_arguments", {})

get_history_arguments = {
    "tags": ["History"],
    "operation_id": "get_history",
    **plural_get_arguments,
}
delete_history_arguments = {
    "tags": ["History"],
    "operation_id": "delete_history",
    **plural_delete_arguments,
}


@resource.register(
    name="history",
    plural_path="/buckets/{{bucket_id}}/history",
    object_path=None,
    plural_methods=("GET", "DELETE"),
    default_arguments={"tags": ["History"], **ViewSet.default_arguments},
    plural_get_arguments=get_history_arguments,
    plural_delete_arguments=delete_history_arguments,
)
class History(resource.Resource):

    schema = HistorySchema

    def get_parent_id(self, request):
        self.bucket_id = request.matchdict["bucket_id"]
        return instance_uri(request, "bucket", id=self.bucket_id)

    def _extract_filters(self):
        filters = super()._extract_filters()
        filters_str_id = []
        for filt in filters:
            if filt.field in ("record_id", "collection_id", "bucket_id"):
                if isinstance(filt.value, int):
                    filt = Filter(filt.field, str(filt.value), filt.operator)
            filters_str_id.append(filt)

        return filters_str_id



from .support import (
    MINIMALIST_BUCKET,
    MINIMALIST_COLLECTION,
    MINIMALIST_GROUP,
    MINIMALIST_RECORD,
    OpenAPITest,
)


class OpenAPIResourcesTest(OpenAPITest):

    allowed_failures = ["version"]

    def test_resource_utilities(self):
        resource = self.resources["Utilities"]
        for op_id, op in resource.operations.items():
            if op_id not in self.allowed_failures:
                self.validate_request_call(op)

    def test_resource_batch(self):
        resource = self.resources["Batch"]
        for op in resource.operations.values():
            requests = [{"path": "/v1/buckets"}]
            defaults = {"method": "POST"}
            self.request.json = lambda: dict(requests=requests, defaults=defaults)
            self.validate_request_call(op)

    def test_resource_buckets(self):
        resource = self.resources["Buckets"]

        for op in resource.operations.values():
            self.app.put_json("/buckets/b1", MINIMALIST_BUCKET, headers=self.headers)
            self.request.path = {"id": "b1"}
            bucket = {**MINIMALIST_BUCKET, "data": {"foo": "bar"}}
            self.request.json = lambda: bucket
            self.validate_request_call(op)

    def test_resource_groups(self):
        resource = self.resources["Groups"]

        for op in resource.operations.values():
            self.app.put_json("/buckets/b1/groups/g1", MINIMALIST_GROUP, headers=self.headers)
            self.request.path = {"bucket_id": "b1", "id": "g1"}
            self.request.json = lambda: MINIMALIST_GROUP
            self.validate_request_call(op)

    def test_resource_collections(self):
        resource = self.resources["Collections"]

        for op in resource.operations.values():
            self.app.put_json(
                "/buckets/b1/collections/c1", MINIMALIST_COLLECTION, headers=self.headers
            )
            self.request.path = {"bucket_id": "b1", "id": "c1"}
            collection = {**MINIMALIST_COLLECTION, "data": {"foo": "bar"}}
            self.request.json = lambda: collection
            self.validate_request_call(op)

    def test_resource_records(self):
        resource = self.resources["Records"]

        for op in resource.operations.values():
            self.app.put_json(
                "/buckets/b1/collections/c1/records/r1", MINIMALIST_RECORD, headers=self.headers
            )
            self.request.path = {"bucket_id": "b1", "collection_id": "c1", "id": "r1"}
            self.request.json = lambda: MINIMALIST_RECORD
            self.validate_request_call(op)



import unittest
from unittest import mock

from kinto.core import scripts


class InitSchemaTest(unittest.TestCase):
    def setUp(self):
        self.registry = mock.MagicMock()

    def test_migrate_calls_initialize_schema_on_backends(self):
        scripts.migrate({"registry": self.registry})
        self.assertTrue(self.registry.storage.initialize_schema.called)
        self.assertTrue(self.registry.cache.initialize_schema.called)
        self.assertTrue(self.registry.permission.initialize_schema.called)

    def test_migrate_skips_missing_backends(self):
        class FakeRegistry:
            settings = dict()
            storage = mock.MagicMock()

        registry = FakeRegistry()
        scripts.migrate({"registry": registry})
        self.assertTrue(registry.storage.initialize_schema.called)

    def test_migrate_in_read_only_display_an_error(self):
        with mock.patch("kinto.core.scripts.logger") as mocked:
            self.registry.settings = {"readonly": "true"}
            scripts.migrate({"registry": self.registry})
            mocked.error.assert_any_call(
                "Cannot migrate the storage backend " "while in readonly mode."
            )
            mocked.error.assert_any_call(
                "Cannot migrate the permission " "backend while in readonly mode."
            )

    def test_migrate_in_dry_run_mode(self):
        scripts.migrate({"registry": self.registry}, dry_run=True)
        reg = self.registry
        reg.storage.initialize_schema.assert_called_with(dry_run=True)
        reg.cache.initialize_schema.assert_called_with(dry_run=True)
        reg.permission.initialize_schema.assert_called_with(dry_run=True)

    def test_flush_cache_clear_the_cache_backend(self):
        scripts.flush_cache({"registry": self.registry})
        reg = self.registry
        reg.cache.flush.assert_called_with()



import logging
import os
import warnings
from collections import defaultdict

from kinto.core.decorators import deprecate_kwargs
from kinto.core.storage import (
    DEFAULT_DELETED_FIELD,
    DEFAULT_ID_FIELD,
    DEFAULT_MODIFIED_FIELD,
    MISSING,
    StorageBase,
    exceptions,
)
from kinto.core.storage.postgresql.client import create_from_config
from kinto.core.storage.postgresql.migrator import MigratorMixin
from kinto.core.utils import COMPARISON

logger = logging.getLogger(__name__)
HERE = os.path.dirname(__file__)


class Storage(StorageBase, MigratorMixin):
    """Storage backend using PostgreSQL.

    Recommended in production (*requires PostgreSQL 9.4 or higher*).

    Enable in configuration::

        kinto.storage_backend = kinto.core.storage.postgresql

    Database location URI can be customized::

        kinto.storage_url = postgresql://user:pass@db.server.lan:5432/dbname

    Alternatively, username and password could also rely on system user ident
    or even specified in :file:`~/.pgpass` (*see PostgreSQL documentation*).

    .. note::

        Some tables and indices are created when ``kinto migrate`` is run.
        This requires some privileges on the database, or some error will
        be raised.

        **Alternatively**, the schema can be initialized outside the
        python application, using the SQL file located in
        :file:`kinto/core/storage/postgresql/schema.sql`. This allows to
        distinguish schema manipulation privileges from schema usage.


    A connection pool is enabled by default::

        kinto.storage_pool_size = 10
        kinto.storage_maxoverflow = 10
        kinto.storage_max_backlog = -1
        kinto.storage_pool_recycle = -1
        kinto.storage_pool_timeout = 30
        kinto.cache_poolclass =
            kinto.core.storage.postgresql.pool.QueuePoolWithMaxBacklog

    The ``max_backlog``  limits the number of threads that can be in the queue
    waiting for a connection.  Once this limit has been reached, any further
    attempts to acquire a connection will be rejected immediately, instead of
    locking up all threads by keeping them waiting in the queue.

    See `dedicated section in SQLAlchemy documentation
    <http://docs.sqlalchemy.org/en/rel_1_0/core/engines.html>`_
    for default values and behaviour.

    .. note::

        Using a `dedicated connection pool <http://pgpool.net>`_ is still
        recommended to allow load balancing, replication or limit the number
        of connections used in a multi-process deployment.

    """  # NOQA

    # MigratorMixin attributes.
    name = "storage"
    schema_version = 21
    schema_file = os.path.join(HERE, "schema.sql")
    migrations_directory = os.path.join(HERE, "migrations")

    def __init__(self, client, max_fetch_size, *args, readonly=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.client = client
        self._max_fetch_size = max_fetch_size
        self.readonly = readonly

    def create_schema(self, dry_run=False):
        """Override create_schema to ensure DB encoding and TZ are OK.
        """
        self._check_database_encoding()
        self._check_database_timezone()
        return super().create_schema(dry_run)

    def initialize_schema(self, dry_run=False):
        return self.create_or_migrate_schema(dry_run)

    def _check_database_timezone(self):
        # Make sure database has UTC timezone.
        query = "SELECT current_setting('TIMEZONE') AS timezone;"
        with self.client.connect() as conn:
            result = conn.execute(query)
            obj = result.fetchone()
        timezone = obj["timezone"].upper()
        if timezone != "UTC":  # pragma: no cover
            msg = f"Database timezone is not UTC ({timezone})"
            warnings.warn(msg)
            logger.warning(msg)

    def _check_database_encoding(self):
        # Make sure database is UTF-8.
        query = """
        SELECT pg_encoding_to_char(encoding) AS encoding
          FROM pg_database
         WHERE datname =  current_database();
        """
        with self.client.connect() as conn:
            result = conn.execute(query)
            obj = result.fetchone()
        encoding = obj["encoding"].lower()
        if encoding != "utf8":  # pragma: no cover
            raise AssertionError(f"Unexpected database encoding {encoding}")

    def get_installed_version(self):
        """Return current version of schema or None if not any found.
        """
        # Check for objects table, which definitely indicates a new
        # DB. (metadata can exist if the permission schema ran first.)
        table_exists_query = """
        SELECT table_name
          FROM information_schema.tables
         WHERE table_name = '{}';
        """
        schema_version_metadata_query = """
        SELECT value AS version
          FROM metadata
         WHERE name = 'storage_schema_version'
         ORDER BY LPAD(value, 3, '0') DESC;
        """
        with self.client.connect() as conn:
            result = conn.execute(table_exists_query.format("objects"))
            objects_table_exists = result.rowcount > 0
            result = conn.execute(table_exists_query.format("records"))
            records_table_exists = result.rowcount > 0

            if not objects_table_exists and not records_table_exists:
                return

            result = conn.execute(schema_version_metadata_query)
            if result.rowcount > 0:
                return int(result.fetchone()["version"])

            # No storage_schema_version row.
            # Perhaps it got flush()ed by a pre-8.1.2 Kinto (which
            # would wipe the metadata table).
            # Alternately, maybe we are working from a very early
            # Cliquet version which never had a migration.
            # Check for a created_at row. If this is gone, it's
            # probably been flushed at some point.
            query = "SELECT COUNT(*) FROM metadata WHERE name = 'created_at';"
            result = conn.execute(query)
            was_flushed = int(result.fetchone()[0]) == 0
            if not was_flushed:
                error_msg = "No schema history; assuming migration from Cliquet (version 1)."
                logger.warning(error_msg)
                return 1

            # We have no idea what the schema is here. Migration
            # is completely broken.
            # Log an obsequious error message to the user and try
            # to recover by assuming the last version where we had
            # this bug.
            logger.warning(UNKNOWN_SCHEMA_VERSION_MESSAGE)

            # This is the last schema version where flushing the
            # server would delete the schema version.
            MAX_FLUSHABLE_SCHEMA_VERSION = 20
            return MAX_FLUSHABLE_SCHEMA_VERSION

    def flush(self):
        """Delete objects from tables without destroying schema.

        This is used in test suites as well as in the flush plugin.
        """
        query = """
        DELETE FROM objects;
        DELETE FROM timestamps;
        """
        with self.client.connect(force_commit=True) as conn:
            conn.execute(query)
        logger.debug("Flushed PostgreSQL storage tables")

    def resource_timestamp(self, resource_name, parent_id):
        query_existing = """
        WITH existing_timestamps AS (
          -- Timestamp of latest object.
          (
            SELECT last_modified, as_epoch(last_modified) AS last_epoch
            FROM objects
            WHERE parent_id = :parent_id
              AND resource_name = :resource_name
            ORDER BY last_modified DESC
            LIMIT 1
          )
          -- Timestamp of empty resource.
          UNION
          (
            SELECT last_modified, as_epoch(last_modified) AS last_epoch
            FROM timestamps
            WHERE parent_id = :parent_id
              AND resource_name = :resource_name
          )
        )
        SELECT MAX(last_modified) AS last_modified, MAX(last_epoch) AS last_epoch
          FROM existing_timestamps
        """

        create_if_missing = """
        INSERT INTO timestamps (parent_id, resource_name, last_modified)
        VALUES (:parent_id, :resource_name, COALESCE(:last_modified, clock_timestamp()::timestamp))
        ON CONFLICT (parent_id, resource_name) DO NOTHING
        RETURNING as_epoch(last_modified) AS last_epoch
        """

        placeholders = dict(parent_id=parent_id, resource_name=resource_name)
        with self.client.connect(readonly=False) as conn:
            existing_ts = None
            ts_result = conn.execute(query_existing, placeholders)
            row = ts_result.fetchone()  # Will return (None, None) when empty.
            existing_ts = row["last_modified"]

            # If the backend is readonly, we should not try to create the timestamp.
            if self.readonly:
                if existing_ts is None:
                    error_msg = (
                        "Cannot initialize empty resource timestamp " "when running in readonly."
                    )
                    raise exceptions.BackendError(message=error_msg)
                obj = row
            else:
                create_result = conn.execute(
                    create_if_missing, dict(last_modified=existing_ts, **placeholders)
                )
                obj = create_result.fetchone() or row

        return obj["last_epoch"]

    @deprecate_kwargs({"collection_id": "resource_name", "record": "obj"})
    def create(
        self,
        resource_name,
        parent_id,
        obj,
        id_generator=None,
        id_field=DEFAULT_ID_FIELD,
        modified_field=DEFAULT_MODIFIED_FIELD,
    ):
        id_generator = id_generator or self.id_generator
        obj = {**obj}
        if id_field in obj:
            # Optimistically raise unicity error if object with same
            # id already exists.
            # Even if this check doesn't find one, be robust against
            # conflicts because we could race with another thread.
            # Still, this reduces write load because SELECTs are
            # cheaper than INSERTs.
            try:
                existing = self.get(resource_name, parent_id, obj[id_field])
                raise exceptions.UnicityError(id_field, existing)
            except exceptions.ObjectNotFoundError:
                pass
        else:
            obj[id_field] = id_generator()

        # Remove redundancy in data field
        query_object = {**obj}
        query_object.pop(id_field, None)
        query_object.pop(modified_field, None)

        # If there is an object in the table and it is deleted = TRUE,
        # we want to replace it. Otherwise, we want to do nothing and
        # throw a UnicityError. Per
        # https://stackoverflow.com/questions/15939902/is-select-or-insert-in-a-function-prone-to-race-conditions/15950324#15950324
        # a WHERE clause in the DO UPDATE will lock the conflicting
        # row whether it is true or not, so the subsequent SELECT is
        # safe. We add a constant "inserted" field to know whether we
        # need to throw or not.
        query = """
        INSERT INTO objects (id, parent_id, resource_name, data, last_modified, deleted)
        VALUES (:object_id, :parent_id,
                :resource_name, (:data)::JSONB,
                from_epoch(:last_modified),
                FALSE)
        ON CONFLICT (id, parent_id, resource_name) DO UPDATE
        SET last_modified = from_epoch(:last_modified),
            data = (:data)::JSONB,
            deleted = FALSE
        WHERE objects.deleted = TRUE
        RETURNING id, data, as_epoch(last_modified) AS last_modified;
        """

        safe_holders = {}
        placeholders = dict(
            object_id=obj[id_field],
            parent_id=parent_id,
            resource_name=resource_name,
            last_modified=obj.get(modified_field),
            data=self.json.dumps(query_object),
        )
        with self.client.connect() as conn:
            result = conn.execute(query % safe_holders, placeholders)
            inserted = result.fetchone()

        if not inserted:
            raise exceptions.UnicityError(id_field)

        obj[modified_field] = inserted["last_modified"]
        return obj

    @deprecate_kwargs({"collection_id": "resource_name"})
    def get(
        self,
        resource_name,
        parent_id,
        object_id,
        id_field=DEFAULT_ID_FIELD,
        modified_field=DEFAULT_MODIFIED_FIELD,
    ):
        query = """
        SELECT as_epoch(last_modified) AS last_modified, data
          FROM objects
         WHERE id = :object_id
           AND parent_id = :parent_id
           AND resource_name = :resource_name
           AND NOT deleted;
        """
        placeholders = dict(object_id=object_id, parent_id=parent_id, resource_name=resource_name)
        with self.client.connect(readonly=True) as conn:
            result = conn.execute(query, placeholders)
            if result.rowcount == 0:
                raise exceptions.ObjectNotFoundError(object_id)
            else:
                existing = result.fetchone()

        obj = existing["data"]
        obj[id_field] = object_id
        obj[modified_field] = existing["last_modified"]
        return obj

    @deprecate_kwargs({"collection_id": "resource_name", "record": "obj"})
    def update(
        self,
        resource_name,
        parent_id,
        object_id,
        obj,
        id_field=DEFAULT_ID_FIELD,
        modified_field=DEFAULT_MODIFIED_FIELD,
    ):

        # Remove redundancy in data field
        query_object = {**obj}
        query_object.pop(id_field, None)
        query_object.pop(modified_field, None)

        query = """
        INSERT INTO objects (id, parent_id, resource_name, data, last_modified, deleted)
        VALUES (:object_id, :parent_id,
                :resource_name, (:data)::JSONB,
                from_epoch(:last_modified),
                FALSE)
        ON CONFLICT (id, parent_id, resource_name) DO UPDATE
        SET data = (:data)::JSONB,
            deleted = FALSE,
            last_modified = GREATEST(from_epoch(:last_modified),
                                     EXCLUDED.last_modified)
        RETURNING as_epoch(last_modified) AS last_modified;
        """
        placeholders = dict(
            object_id=object_id,
            parent_id=parent_id,
            resource_name=resource_name,
            last_modified=obj.get(modified_field),
            data=self.json.dumps(query_object),
        )

        with self.client.connect() as conn:
            result = conn.execute(query, placeholders)
            updated = result.fetchone()

        obj = {**obj, id_field: object_id}
        obj[modified_field] = updated["last_modified"]
        return obj

    @deprecate_kwargs({"collection_id": "resource_name"})
    def delete(
        self,
        resource_name,
        parent_id,
        object_id,
        id_field=DEFAULT_ID_FIELD,
        with_deleted=True,
        modified_field=DEFAULT_MODIFIED_FIELD,
        deleted_field=DEFAULT_DELETED_FIELD,
        last_modified=None,
    ):
        if with_deleted:
            query = """
            UPDATE objects
               SET deleted=TRUE,
                   data=(:deleted_data)::JSONB,
                   last_modified=from_epoch(:last_modified)
             WHERE id = :object_id
               AND parent_id = :parent_id
               AND resource_name = :resource_name
               AND NOT deleted
            RETURNING as_epoch(last_modified) AS last_modified;
            """
        else:
            query = """
            DELETE FROM objects
            WHERE id = :object_id
               AND parent_id = :parent_id
               AND resource_name = :resource_name
               AND NOT deleted
            RETURNING as_epoch(last_modified) AS last_modified;
            """
        deleted_data = self.json.dumps(dict([(deleted_field, True)]))
        placeholders = dict(
            object_id=object_id,
            parent_id=parent_id,
            resource_name=resource_name,
            last_modified=last_modified,
            deleted_data=deleted_data,
        )

        with self.client.connect() as conn:
            result = conn.execute(query, placeholders)
            if result.rowcount == 0:
                raise exceptions.ObjectNotFoundError(object_id)
            updated = result.fetchone()

        obj = {}
        obj[modified_field] = updated["last_modified"]
        obj[id_field] = object_id

        obj[deleted_field] = True
        return obj

    @deprecate_kwargs({"collection_id": "resource_name"})
    def delete_all(
        self,
        resource_name,
        parent_id,
        filters=None,
        sorting=None,
        pagination_rules=None,
        limit=None,
        id_field=DEFAULT_ID_FIELD,
        with_deleted=True,
        modified_field=DEFAULT_MODIFIED_FIELD,
        deleted_field=DEFAULT_DELETED_FIELD,
    ):
        if with_deleted:
            query = """
            WITH matching_objects AS (
                SELECT id, parent_id, resource_name
                    FROM objects
                    WHERE {parent_id_filter}
                          {resource_name_filter}
                          AND NOT deleted
                          {conditions_filter}
                          {pagination_rules}
                    {sorting}
                    LIMIT :pagination_limit
                    FOR UPDATE
            )
            UPDATE objects
               SET deleted=TRUE, data=(:deleted_data)::JSONB, last_modified=NULL
              FROM matching_objects
             WHERE objects.id = matching_objects.id
               AND objects.parent_id = matching_objects.parent_id
               AND objects.resource_name = matching_objects.resource_name
            RETURNING objects.id, as_epoch(last_modified) AS last_modified;
            """
        else:
            query = """
            WITH matching_objects AS (
                SELECT id, parent_id, resource_name
                    FROM objects
                    WHERE {parent_id_filter}
                          {resource_name_filter}
                          AND NOT deleted
                          {conditions_filter}
                          {pagination_rules}
                    {sorting}
                    LIMIT :pagination_limit
                    FOR UPDATE
            )
            DELETE
            FROM objects
            USING matching_objects
            WHERE objects.id = matching_objects.id
              AND objects.parent_id = matching_objects.parent_id
              AND objects.resource_name = matching_objects.resource_name
            RETURNING objects.id, as_epoch(last_modified) AS last_modified;
            """

        id_field = id_field or self.id_field
        modified_field = modified_field or self.modified_field
        deleted_data = self.json.dumps(dict([(deleted_field, True)]))
        placeholders = dict(
            parent_id=parent_id, resource_name=resource_name, deleted_data=deleted_data
        )
        # Safe strings
        safeholders = defaultdict(str)
        # Handle parent_id as a regex only if it contains *
        if "*" in parent_id:
            safeholders["parent_id_filter"] = "parent_id LIKE :parent_id"
            placeholders["parent_id"] = parent_id.replace("*", "%")
        else:
            safeholders["parent_id_filter"] = "parent_id = :parent_id"
        # If resource is None, remove it from query.
        if resource_name is None:
            safeholders["resource_name_filter"] = ""
        else:
            safeholders["resource_name_filter"] = "AND resource_name = :resource_name"  # NOQA

        if filters:
            safe_sql, holders = self._format_conditions(filters, id_field, modified_field)
            safeholders["conditions_filter"] = f"AND {safe_sql}"
            placeholders.update(**holders)

        if sorting:
            sql, holders = self._format_sorting(sorting, id_field, modified_field)
            safeholders["sorting"] = sql
            placeholders.update(**holders)

        if pagination_rules:
            sql, holders = self._format_pagination(pagination_rules, id_field, modified_field)
            safeholders["pagination_rules"] = f"AND ({sql})"
            placeholders.update(**holders)

        # Limit the number of results (pagination).
        limit = min(self._max_fetch_size, limit) if limit else self._max_fetch_size
        placeholders["pagination_limit"] = limit

        with self.client.connect() as conn:
            result = conn.execute(query.format_map(safeholders), placeholders)
            deleted = result.fetchmany(self._max_fetch_size)

        objects = []
        for result in deleted:
            obj = {}
            obj[id_field] = result["id"]
            obj[modified_field] = result["last_modified"]
            obj[deleted_field] = True
            objects.append(obj)

        return objects

    @deprecate_kwargs({"collection_id": "resource_name"})
    def purge_deleted(
        self,
        resource_name,
        parent_id,
        before=None,
        id_field=DEFAULT_ID_FIELD,
        modified_field=DEFAULT_MODIFIED_FIELD,
    ):
        delete_tombstones = """
        DELETE
        FROM objects
        WHERE {parent_id_filter}
              {resource_name_filter}
              {conditions_filter}
        """
        id_field = id_field or self.id_field
        modified_field = modified_field or self.modified_field
        placeholders = dict(parent_id=parent_id, resource_name=resource_name)
        # Safe strings
        safeholders = defaultdict(str)
        # Handle parent_id as a regex only if it contains *
        if "*" in parent_id:
            safeholders["parent_id_filter"] = "parent_id LIKE :parent_id"
            placeholders["parent_id"] = parent_id.replace("*", "%")
        else:
            safeholders["parent_id_filter"] = "parent_id = :parent_id"
        # If resource is None, remove it from query.
        if resource_name is None:
            safeholders["resource_name_filter"] = ""
        else:
            safeholders["resource_name_filter"] = "AND resource_name = :resource_name"  # NOQA

        if before is not None:
            safeholders["conditions_filter"] = "AND as_epoch(last_modified) < :before"
            placeholders["before"] = before

        with self.client.connect() as conn:
            result = conn.execute(delete_tombstones.format_map(safeholders), placeholders)
            deleted = result.rowcount

            # If purging everything from a parent_id, then clear timestamps.
            if resource_name is None and before is None:
                delete_timestamps = """
                DELETE
                FROM timestamps
                WHERE {parent_id_filter}
                """
                conn.execute(delete_timestamps.format_map(safeholders), placeholders)

        return deleted

    def list_all(
        self,
        resource_name,
        parent_id,
        filters=None,
        sorting=None,
        pagination_rules=None,
        limit=None,
        include_deleted=False,
        id_field=DEFAULT_ID_FIELD,
        modified_field=DEFAULT_MODIFIED_FIELD,
        deleted_field=DEFAULT_DELETED_FIELD,
    ):

        query = """
            SELECT id, as_epoch(last_modified) AS last_modified, data
            FROM objects
            WHERE {parent_id_filter}
            AND resource_name = :resource_name
            {conditions_deleted}
            {conditions_filter}
            {pagination_rules}
            {sorting}
            LIMIT :pagination_limit;
        """

        rows = self._get_rows(
            query,
            resource_name,
            parent_id,
            filters=filters,
            sorting=sorting,
            pagination_rules=pagination_rules,
            limit=limit,
            include_deleted=include_deleted,
            id_field=id_field,
            modified_field=modified_field,
            deleted_field=deleted_field,
        )

        if len(rows) == 0:
            return []

        records = []
        for result in rows:
            record = result["data"]
            record[id_field] = result["id"]
            record[modified_field] = result["last_modified"]
            records.append(record)
        return records

    def count_all(
        self,
        resource_name,
        parent_id,
        filters=None,
        id_field=DEFAULT_ID_FIELD,
        modified_field=DEFAULT_MODIFIED_FIELD,
        deleted_field=DEFAULT_DELETED_FIELD,
    ):

        query = """
            SELECT COUNT(*) AS total_count
            FROM objects
            WHERE {parent_id_filter}
            AND resource_name = :resource_name
            AND NOT deleted
            {conditions_filter}
        """
        rows = self._get_rows(
            query,
            resource_name,
            parent_id,
            filters=filters,
            id_field=id_field,
            modified_field=modified_field,
            deleted_field=deleted_field,
        )
        return rows[0]["total_count"]

    def _get_rows(
        self,
        query,
        resource_name,
        parent_id,
        filters=None,
        sorting=None,
        pagination_rules=None,
        limit=None,
        include_deleted=False,
        id_field=DEFAULT_ID_FIELD,
        modified_field=DEFAULT_MODIFIED_FIELD,
        deleted_field=DEFAULT_DELETED_FIELD,
    ):

        # Unsafe strings escaped by PostgreSQL
        placeholders = dict(parent_id=parent_id, resource_name=resource_name)

        # Safe strings
        safeholders = defaultdict(str)

        # Handle parent_id as a regex only if it contains *
        if "*" in parent_id:
            safeholders["parent_id_filter"] = "parent_id LIKE :parent_id"
            placeholders["parent_id"] = parent_id.replace("*", "%")
        else:
            safeholders["parent_id_filter"] = "parent_id = :parent_id"

        if filters:
            safe_sql, holders = self._format_conditions(filters, id_field, modified_field)
            safeholders["conditions_filter"] = f"AND {safe_sql}"
            placeholders.update(**holders)

        if not include_deleted:
            safeholders["conditions_deleted"] = "AND NOT deleted"

        if sorting:
            sql, holders = self._format_sorting(sorting, id_field, modified_field)
            safeholders["sorting"] = sql
            placeholders.update(**holders)

        if pagination_rules:
            sql, holders = self._format_pagination(pagination_rules, id_field, modified_field)
            safeholders["pagination_rules"] = f"AND ({sql})"
            placeholders.update(**holders)

        # Limit the number of results (pagination).
        limit = min(self._max_fetch_size + 1, limit) if limit else self._max_fetch_size
        placeholders["pagination_limit"] = limit

        with self.client.connect(readonly=True) as conn:
            result = conn.execute(query.format_map(safeholders), placeholders)
            return result.fetchmany(self._max_fetch_size + 1)

    def _format_conditions(self, filters, id_field, modified_field, prefix="filters"):
        """Format the filters list in SQL, with placeholders for safe escaping.

        .. note::
            All conditions are combined using AND.

        .. note::

            Field name and value are escaped as they come from HTTP API.

        :returns: A SQL string with placeholders, and a dict mapping
            placeholders to actual values.
        :rtype: tuple
        """
        operators = {
            COMPARISON.EQ: "=",
            COMPARISON.NOT: "<>",
            COMPARISON.IN: "IN",
            COMPARISON.EXCLUDE: "NOT IN",
            COMPARISON.LIKE: "ILIKE",
            COMPARISON.CONTAINS: "@>",
        }

        conditions = []
        holders = {}
        for i, filtr in enumerate(filters):
            value = filtr.value
            is_like_query = filtr.operator == COMPARISON.LIKE

            if filtr.field == id_field:
                sql_field = "id"
                if isinstance(value, int):
                    value = str(value)
            elif filtr.field == modified_field:
                sql_field = "as_epoch(last_modified)"
            else:
                column_name = "data"
                # Subfields: ``person.name`` becomes ``data->person->>name``
                subfields = filtr.field.split(".")
                for j, subfield in enumerate(subfields):
                    # Safely escape field name
                    field_holder = f"{prefix}_field_{i}_{j}"
                    holders[field_holder] = subfield
                    # Use ->> to convert the last level to text if
                    # needed for LIKE query. (Other queries do JSONB comparison.)
                    column_name += "->>" if j == len(subfields) - 1 and is_like_query else "->"
                    column_name += f":{field_holder}"
                sql_field = column_name

            string_field = filtr.field in (id_field, modified_field) or is_like_query
            if not string_field and value != MISSING:
                # JSONB-ify the value.
                if filtr.operator not in (
                    COMPARISON.IN,
                    COMPARISON.EXCLUDE,
                    COMPARISON.CONTAINS_ANY,
                ):
                    value = self.json.dumps(value)
                else:
                    value = [self.json.dumps(v) for v in value]

            if filtr.operator in (COMPARISON.IN, COMPARISON.EXCLUDE):
                value = tuple(value)
                # WHERE field IN ();  -- Fails with syntax error.
                if len(value) == 0:
                    value = (None,)

            if is_like_query:
                # Operand should be a string.
                # Add implicit start/end wildcards if none is specified.
                if "*" not in value:
                    value = f"*{value}*"
                value = value.replace("*", "%")

            if filtr.operator == COMPARISON.HAS:
                operator = "IS NOT NULL" if filtr.value else "IS NULL"
                cond = f"{sql_field} {operator}"

            elif filtr.operator == COMPARISON.CONTAINS_ANY:
                value_holder = f"{prefix}_value_{i}"
                holders[value_holder] = value
                # In case the field is not a sequence, we ignore the object.
                is_json_sequence = f"jsonb_typeof({sql_field}) = 'array'"
                # Postgres's && operator doesn't support jsonbs.
                # However, it does support Postgres arrays of any
                # type. Assume that the referenced field is a JSON
                # array and convert it to a Postgres array.
                data_as_array = f"""
                (SELECT array_agg(elems) FROM jsonb_array_elements({sql_field}) elems)
                """
                cond = f"{is_json_sequence} AND {data_as_array} && (:{value_holder})::jsonb[]"

            elif value != MISSING:
                # Safely escape value. MISSINGs get handled below.
                value_holder = f"{prefix}_value_{i}"
                holders[value_holder] = value

                sql_operator = operators.setdefault(filtr.operator, filtr.operator.value)
                cond = f"{sql_field} {sql_operator} :{value_holder}"

            # If the field is missing, column_name will produce
            # NULL. NULL has strange properties with comparisons
            # in SQL -- NULL = anything => NULL, NULL <> anything => NULL.
            # We generally want missing fields to be treated as a
            # special value that compares as different from
            # everything, including JSON null. Do this on a
            # per-operator basis.
            null_false_operators = (
                # NULLs aren't EQ to anything (definitionally).
                COMPARISON.EQ,
                # So they can't match anything in an INCLUDE.
                COMPARISON.IN,
                # Nor can they be LIKE anything.
                COMPARISON.LIKE,
                # NULLs don't contain anything.
                COMPARISON.CONTAINS,
                COMPARISON.CONTAINS_ANY,
            )
            null_true_operators = (
                # NULLs are automatically not equal to everything.
                COMPARISON.NOT,
                # Thus they can never be excluded.
                COMPARISON.EXCLUDE,
                # Match Postgres's default sort behavior
                # (NULLS LAST) by allowing NULLs to
                # automatically be greater than everything.
                COMPARISON.GT,
                COMPARISON.MIN,
            )

            if not (filtr.field == id_field or filtr.field == modified_field):
                if value == MISSING:
                    # Handle MISSING values. The main use case for this is
                    # pagination, since there's no way to encode MISSING
                    # at the HTTP API level. Because we only need to cover
                    # pagination, we don't have to worry about any
                    # operators besides LT, LE, GT, GE, and EQ, and
                    # never worry about id_field or modified_field.
                    #
                    # Comparing a value against NULL is not the same
                    # as comparing a NULL against some other value, so
                    # we need another set of operators for which
                    # NULLs are OK.
                    if filtr.operator in (COMPARISON.EQ, COMPARISON.MIN):
                        # If a row is NULL, then it can be == NULL
                        # (for the purposes of pagination).
                        # >= NULL should only match rows that are
                        # NULL, since there's nothing higher.
                        cond = f"{sql_field} IS NULL"
                    elif filtr.operator == COMPARISON.LT:
                        # If we're looking for < NULL, match only
                        # non-nulls.
                        cond = f"{sql_field} IS NOT NULL"
                    elif filtr.operator == COMPARISON.MAX:
                        # <= NULL should include everything -- NULL
                        # because it's equal, and non-nulls because
                        # they're <.
                        cond = "TRUE"
                    elif filtr.operator == COMPARISON.GT:
                        # Nothing can be greater than NULL (that is,
                        # higher in search order).
                        cond = "FALSE"
                    else:
                        raise ValueError("Somehow we got a filter with MISSING value")
                elif filtr.operator in null_false_operators:
                    cond = f"({sql_field} IS NOT NULL AND {cond})"
                elif filtr.operator in null_true_operators:
                    cond = f"({sql_field} IS NULL OR {cond})"
                else:
                    # No need to check for LT and MAX because NULL < foo
                    # is NULL, which is falsy in SQL.
                    pass

            conditions.append(cond)

        safe_sql = " AND ".join(conditions)
        return safe_sql, holders

    def _format_pagination(self, pagination_rules, id_field, modified_field):
        """Format the pagination rules in SQL, with placeholders for
        safe escaping.

        .. note::

            All rules are combined using OR.

        .. note::

            Field names are escaped as they come from HTTP API.

        :returns: A SQL string with placeholders, and a dict mapping
            placeholders to actual values.
        :rtype: tuple
        """
        rules = []
        placeholders = {}

        for i, rule in enumerate(pagination_rules):
            prefix = f"rules_{i}"
            safe_sql, holders = self._format_conditions(
                rule, id_field, modified_field, prefix=prefix
            )
            rules.append(safe_sql)
            placeholders.update(**holders)

        # Unsure how to convert to fstrings
        safe_sql = " OR ".join([f"({r})" for r in rules])
        return safe_sql, placeholders

    def _format_sorting(self, sorting, id_field, modified_field):
        """Format the sorting in SQL, with placeholders for safe escaping.

        .. note::

            Field names are escaped as they come from HTTP API.

        :returns: A SQL string with placeholders, and a dict mapping
            placeholders to actual values.
        :rtype: tuple
        """
        sorts = []
        holders = {}
        for i, sort in enumerate(sorting):

            if sort.field == id_field:
                sql_field = "id"
            elif sort.field == modified_field:
                sql_field = "last_modified"
            else:
                # Subfields: ``person.name`` becomes ``data->person->name``
                subfields = sort.field.split(".")
                sql_field = "data"
                for j, subfield in enumerate(subfields):
                    # Safely escape field name
                    field_holder = f"sort_field_{i}_{j}"
                    holders[field_holder] = subfield
                    sql_field += f"->(:{field_holder})"

            sql_direction = "ASC" if sort.direction > 0 else "DESC"
            sql_sort = f"{sql_field} {sql_direction}"
            sorts.append(sql_sort)

        safe_sql = f"ORDER BY {', '.join(sorts)}"
        return safe_sql, holders


def load_from_config(config):
    settings = config.get_settings()
    max_fetch_size = int(settings["storage_max_fetch_size"])
    strict = settings.get("storage_strict_json", False)
    readonly = settings.get("readonly", False)
    client = create_from_config(config, prefix="storage_")
    return Storage(
        client=client, max_fetch_size=max_fetch_size, strict_json=strict, readonly=readonly
    )


UNKNOWN_SCHEMA_VERSION_MESSAGE = """
Missing schema history. Perhaps at some point, this Kinto server was
flushed.  Due to a bug in older Kinto versions (see
https://github.com/Kinto/kinto/issues/1460), flushing the server would
cause us to forget what version of the schema was in use. This means
automatic migration is impossible.

Historically, when this happened, Kinto would just assume that the
wiped server had the "current" schema, so you may have been missing a
schema version for quite some time.

To try to recover, we have assumed a schema version corresponding to
the last Kinto version with this bug (schema version 20). However, if
a migration fails, or most queries are broken, you may not actually be
running that schema. You can try to fix this by manually setting the
schema version in the database to what you think it should be using a
command like:

    INSERT INTO metadata VALUES ('storage_schema_version', '19');

See https://github.com/Kinto/kinto/wiki/Schema-versions for more details.

""".strip()



from unittest import TestCase, mock

from kinto.core.resource.model import ShareableModel

from . import BaseTest


class ModelTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.obj = self.model.create_object({"field": "value"})
        self.obj.pop(self.model.permissions_field)
        self.resource.model.get_permission_object_id = lambda x: "/object/id"

    def test_list_returns_all_objects_in_data(self):
        result = self.resource.plural_get()
        objects = result["data"]
        self.assertEqual(len(objects), 1)
        self.assertDictEqual(objects[0], self.obj)


class CreateTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.resource.request.validated["body"] = {"data": {"field": "new"}}

    def test_new_objects_are_linked_to_owner(self):
        resp = self.resource.plural_post()["data"]
        object_id = resp["id"]
        self.model.get_object(object_id)  # not raising

    def test_create_object_returns_at_least_id_and_last_modified(self):
        obj = self.resource.plural_post()["data"]
        self.assertIn(self.resource.model.id_field, obj)
        self.assertIn(self.resource.model.modified_field, obj)
        self.assertIn("field", obj)


class DeleteModelTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.patch_known_field.start()
        self.model.create_object({"field": "a"})
        self.model.create_object({"field": "b"})

    def test_delete_on_list_removes_all_objects(self):
        self.resource.plural_delete()
        result = self.resource.plural_get()
        objects = result["data"]
        self.assertEqual(len(objects), 0)

    def test_delete_returns_deleted_version_of_objects(self):
        result = self.resource.plural_delete()
        deleted = result["data"][0]
        self.assertIn("deleted", deleted)

    def test_plural_delete_supports_filters(self):
        self.resource.request.validated["querystring"] = {"field": "a"}
        self.resource.plural_delete()
        self.resource.request.validated["querystring"] = {}
        result = self.resource.plural_get()
        objects = result["data"]
        self.assertEqual(len(objects), 1)


class DeprecatedMethodsTest(BaseTest):
    def setUp(self):
        super().setUp()
        patch = mock.patch("warnings.warn")
        self.mocked_warnings = patch.start()
        # BaseTest will stops all patches in tearDown().

    def test_collection_id(self):
        self.resource.model.collection_id

        message = "`collection_id` is deprecated, use `resource_name` instead."
        self.mocked_warnings.assert_called_with(message, DeprecationWarning)

    def test_get_records(self, *args, **kwargs):
        self.resource.model.get_records()

        message = "`get_records()` is deprecated, use `get_objects()` instead."
        self.mocked_warnings.assert_called_with(message, DeprecationWarning)

    def test_delete_records(self, *args, **kwargs):
        self.resource.model.delete_records()

        message = "`delete_records()` is deprecated, use `delete_objects()` instead."
        self.mocked_warnings.assert_called_with(message, DeprecationWarning)

    def test_get_record(self, *args, **kwargs):
        self.resource.model.create_object({"id": "abc"})

        self.resource.model.get_record(record_id="abc")

        message = "`get_record()` is deprecated, use `get_object()` instead."
        self.mocked_warnings.assert_called_with(message, DeprecationWarning)

    def test_create_record(self, *args, **kwargs):
        self.resource.model.create_record({"id": "abc"})

        message = "`create_record()` is deprecated, use `create_object()` instead."
        self.mocked_warnings.assert_called_with(message, DeprecationWarning)

    def test_update_record(self, *args, **kwargs):
        self.resource.model.update_record({"id": "abc"})

        message = "`update_record()` is deprecated, use `update_object()` instead."
        self.mocked_warnings.assert_called_with(message, DeprecationWarning)

    def test_delete_record(self, *args, **kwargs):
        self.resource.model.create_object({"id": "abc"})

        self.resource.model.delete_record(record={"id": "abc"})

        message = "`delete_record()` is deprecated, use `delete_object()` instead."
        self.mocked_warnings.assert_called_with(message, DeprecationWarning)


class DeprecatedShareableModel(TestCase):
    def test_deprecated_warning(self):
        with mock.patch("warnings.warn") as mocked_warnings:
            ShareableModel(storage=mock.MagicMock(), permission=mock.MagicMock())

        message = "`ShareableModel` is deprecated, use `Model` instead."
        mocked_warnings.assert_called_with(message, DeprecationWarning)



import logging
import warnings
from collections import OrderedDict
from enum import Enum

import pyramid.tweens
import transaction
from pyramid.events import NewRequest

from kinto.core.utils import strip_uri_prefix

logger = logging.getLogger(__name__)


class ACTIONS(Enum):
    CREATE = "create"
    DELETE = "delete"
    READ = "read"
    UPDATE = "update"

    @staticmethod
    def from_string_list(elements):
        return tuple(ACTIONS(el) for el in elements)


class _ResourceEvent:
    def __init__(self, payload, request):
        self.payload = payload
        self.request = request

    def __repr__(self):
        return f"<{self.__class__.__name__} action={self.payload['action']} uri={self.payload['uri']}>"

    @property
    def read_records(self):
        message = "`read_records` is deprecated, use `read_objects` instead."
        warnings.warn(message, DeprecationWarning)
        return self.read_objects

    @property
    def impacted_records(self):
        message = "`impacted_records` is deprecated, use `impacted_objects` instead."
        warnings.warn(message, DeprecationWarning)
        return self.impacted_objects


class ResourceRead(_ResourceEvent):
    """Triggered when a resource is being read.
    """

    def __init__(self, payload, read_objects, request):
        super().__init__(payload, request)
        self.read_objects = read_objects


class ResourceChanged(_ResourceEvent):
    """Triggered when a resource is being changed.
    """

    def __init__(self, payload, impacted_objects, request):
        super().__init__(payload, request)
        self.impacted_objects = impacted_objects


class AfterResourceRead(_ResourceEvent):
    """Triggered after a resource was successfully read.
    """

    def __init__(self, payload, read_objects, request):
        super().__init__(payload, request)
        self.read_objects = read_objects


class AfterResourceChanged(_ResourceEvent):
    """Triggered after a resource was successfully changed.
    """

    def __init__(self, payload, impacted_objects, request):
        super().__init__(payload, request)
        self.impacted_objects = impacted_objects


class EventCollector(object):
    """A collection to gather events emitted over the course of a request.

    Events are gathered by parent id, resource type, and event
    type. This serves as a primitive normalization so that we can emit
    fewer events.
    """

    def __init__(self):
        self.event_dict = OrderedDict()
        """The events as collected so far.

        The key of the event_dict is a triple (resource_name,
        parent_id, action). The value is a triple (impacted, request,
        payload). If the same (resource_name, parent_id, action) is
        encountered, we just extend the existing impacted with the new
        impacted. N.B. this means all values in the payload must not
        be specific to a single impacted_object. See
        https://github.com/Kinto/kinto/issues/945 and
        https://github.com/Kinto/kinto/issues/1731.
        """

    def add_event(self, resource_name, parent_id, action, payload, impacted, request):
        key = (resource_name, parent_id, action)
        if key not in self.event_dict:
            value = (payload, impacted, request)
            self.event_dict[key] = value
        else:
            old_value = self.event_dict[key]
            (old_payload, old_impacted, old_request) = old_value
            # May be a good idea to assert that old_payload == payload here.
            self.event_dict[key] = (old_payload, old_impacted + impacted, old_request)

    def drain(self):
        """Return an iterator that removes elements from this EventCollector.

        This can be used to process events while still allowing events
        to be added (for instance, as part of a cascade where events
        add other events).

        Items yielded will be of a tuple suitable for using as
        arguments to EventCollector.add_event.
        """
        return EventCollectorDrain(self)


class EventCollectorDrain(object):
    """An iterator that drains an EventCollector.

    Get one using EventCollector.drain()."""

    def __init__(self, event_collector):
        self.event_collector = event_collector

    def __iter__(self):
        return self

    def __next__(self):
        if self.event_collector.event_dict:
            # Get the "first" key in insertion order, so as to process
            # events in the same order they were queued.
            key = next(iter(self.event_collector.event_dict.keys()))
            value = self.event_collector.event_dict.pop(key)
            return key + value
        else:
            raise StopIteration


def notify_resource_events_before(handler, registry):
    """Tween that runs ResourceChanged events.

    This tween runs under EXCVIEW, so exceptions raised by
    ResourceChanged events are caught and dealt the same as any other
    exceptions.

    """

    def tween(request):
        response = handler(request)
        for event in request.get_resource_events():
            request.registry.notify(event)

        return response

    return tween


def setup_transaction_hook(config):
    """
    Resource events are plugged with the transactions of ``pyramid_tm``.

    Once a transaction is committed, ``AfterResourceRead`` and
    ``AfterResourceChanged`` events are sent.
    """

    def _notify_resource_events_after(success, request):
        """Notify the accumulated resource events if transaction succeeds.
        """
        if not success:  # pragma: no cover
            return

        for event in request.get_resource_events(after_commit=True):
            try:
                request.registry.notify(event)
            except Exception:
                logger.error("Unable to notify", exc_info=True)

    def on_new_request(event):
        """When a new request comes in, hook on transaction commit.
        """
        # Since there is one transaction per batch, ignore subrequests.
        if hasattr(event.request, "parent"):
            return
        current = transaction.get()
        current.addAfterCommitHook(_notify_resource_events_after, args=(event.request,))

    config.add_subscriber(on_new_request, NewRequest)
    config.add_tween(
        "kinto.core.events.notify_resource_events_before", under=pyramid.tweens.EXCVIEW
    )


def get_resource_events(request, after_commit=False):
    """Generator to iterate the list of events triggered on resources.

    The list is sorted chronologically (see OrderedDict).

    This drains the resource_events currently in the request, which
    allows us to process new events as they are added by current
    events. However, once the iteration is over, we merge all the
    events we've emitted into a new resource_events, which we store on
    the request so we can reprocess the same events in an after-commit
    tween.

    This generator must be completely consumed!
    """
    by_resource = request.bound_data.get("resource_events", EventCollector())
    afterwards = EventCollector()

    for event_call in by_resource.drain():
        afterwards.add_event(*event_call)
        (_, _, action, payload, impacted, request) = event_call

        if after_commit:
            if action == ACTIONS.READ:
                event_cls = AfterResourceRead
            else:
                event_cls = AfterResourceChanged
        else:
            if action == ACTIONS.READ:
                event_cls = ResourceRead
            else:
                event_cls = ResourceChanged

        yield event_cls(payload, impacted, request)

    request.bound_data["resource_events"] = afterwards


def notify_resource_event(
    request, parent_id, timestamp, data, action, old=None, resource_name=None, resource_data=None
):
    """Request helper to stack a resource event.

    If a similar event (same resource, same action) already occured during the
    current transaction (e.g. batch) then just extend the impacted objects of
    the previous one.

    :param resource_name: The name of the resource on which the event
        happened (taken from the request if not provided).
    :param resource_data: Information about the resource on which the
        event is being emitted. Usually contains information about how
        to find this object in the hierarchy (for instance,
        ``bucket_id`` and ``collection_id`` for a record). Taken from
        the request matchdict if absent.
    :type resource_data: dict

    """
    if action == ACTIONS.READ:
        if not isinstance(data, list):
            data = [data]
        impacted = data
    elif action == ACTIONS.CREATE:
        impacted = [{"new": data}]
    elif action == ACTIONS.DELETE:
        if not isinstance(data, list):
            impacted = [{"new": data, "old": old}]
        else:
            impacted = []
            for i, new in enumerate(data):
                impacted.append({"new": new, "old": old[i]})
    else:  # ACTIONS.UPDATE:
        impacted = [{"new": data, "old": old}]

    # Get previously triggered events.
    events = request.bound_data.setdefault("resource_events", EventCollector())

    resource_name = resource_name or request.current_resource_name
    matchdict = resource_data or dict(request.matchdict)

    payload = {
        "timestamp": timestamp,
        "action": action.value,
        # Deprecated: don't actually use URI (see #945).
        "uri": strip_uri_prefix(request.path),
        "user_id": request.prefixed_userid,
        "resource_name": resource_name,
    }

    # Deprecated: don't actually use `resource_name_id` either (see #945).
    if "id" in request.matchdict:
        matchdict[resource_name + "_id"] = matchdict.pop("id")

    payload.update(**matchdict)

    events.add_event(resource_name, parent_id, action, payload, impacted, request)



"""Main entry point
"""
import logging
import tempfile

import pkg_resources
from cornice import Service as CorniceService
from dockerflow import logging as dockerflow_logging
from pyramid.settings import aslist

from kinto.core import errors, events
from kinto.core.initialization import (  # NOQA
    initialize,
    install_middlewares,
    load_default_settings,
)
from kinto.core.utils import (
    current_resource_name,
    current_service,
    follow_subrequest,
    log_context,
    prefixed_principals,
    prefixed_userid,
)

logger = logging.getLogger(__name__)

__all__ = ["initialize", "load_default_settings", "Service"]

# Module version, as defined in PEP-0396.
__version__ = pkg_resources.get_distribution("kinto").version  # FIXME?

DEFAULT_SETTINGS = {
    "backoff": None,
    "backoff_percentage": None,
    "batch_max_requests": 25,
    "cache_backend": "",
    "cache_hosts": "",
    "cache_url": "",
    "cache_prefix": "",
    "cache_max_size_bytes": 524288,
    "cors_origins": "*",
    "cors_max_age_seconds": 3600,
    "eos": None,
    "eos_message": None,
    "eos_url": None,
    "error_info_link": "https://github.com/Kinto/kinto/issues/",
    "http_host": None,
    "http_scheme": None,
    "id_generator": "kinto.core.storage.generators.UUID4",
    "includes": "",
    "initialization_sequence": (
        "kinto.core.initialization.setup_request_bound_data",
        "kinto.core.initialization.setup_json_serializer",
        "kinto.core.initialization.setup_csp_headers",
        "kinto.core.initialization.setup_logging",
        "kinto.core.initialization.setup_storage",
        "kinto.core.initialization.setup_permission",
        "kinto.core.initialization.setup_cache",
        "kinto.core.initialization.setup_requests_scheme",
        "kinto.core.initialization.setup_version_redirection",
        "kinto.core.initialization.setup_deprecation",
        "kinto.core.initialization.setup_authentication",
        "kinto.core.initialization.setup_backoff",
        "kinto.core.initialization.setup_statsd",
        "kinto.core.initialization.setup_listeners",
        "kinto.core.events.setup_transaction_hook",
    ),
    "event_listeners": "",
    "heartbeat_timeout_seconds": 10,
    "newrelic_config": None,
    "newrelic_env": "dev",
    "paginate_by": None,
    "pagination_token_validity_seconds": 10 * 60,
    "permission_backend": "",
    "permission_url": "",
    "profiler_dir": tempfile.gettempdir(),
    "profiler_enabled": False,
    "project_docs": "",
    "project_name": "",
    "project_version": "",
    "readonly": False,
    "retry_after_seconds": 30,
    "settings_prefix": "",
    "statsd_backend": "kinto.core.statsd",
    "statsd_prefix": "kinto.core",
    "statsd_url": None,
    "storage_backend": "",
    "storage_url": "",
    "storage_max_fetch_size": 10000,
    "tm.annotate_user": False,  # Do annotate transactions with the user-id.
    "transaction_per_request": True,
    "userid_hmac_secret": "",
    "version_json_path": "version.json",
    "version_prefix_redirect_enabled": True,
    "trailing_slash_redirect_enabled": True,
    "multiauth.groupfinder": "kinto.core.authorization.groupfinder",
    "multiauth.policies": "",
    "multiauth.policy.basicauth.use": (
        "kinto.core.authentication." "BasicAuthAuthenticationPolicy"
    ),
    "multiauth.authorization_policy": ("kinto.core.authorization." "AuthorizationPolicy"),
}


class Service(CorniceService):
    """Subclass of the default cornice service.

    This is useful in order to attach specific behaviours without monkey
    patching the default cornice service (which would impact other uses of it)
    """

    default_cors_headers = ("Backoff", "Retry-After", "Alert", "Content-Length", "Content-Type")

    def error_handler(self, request):
        return errors.json_error_handler(request)

    @classmethod
    def init_from_settings(cls, settings):
        cls.cors_origins = tuple(aslist(settings["cors_origins"]))
        cors_max_age = settings["cors_max_age_seconds"]
        cls.cors_max_age = int(cors_max_age) if cors_max_age else None


class JsonLogFormatter(dockerflow_logging.JsonLogFormatter):
    logger_name = "kinto"

    @classmethod
    def init_from_settings(cls, settings):
        cls.logger_name = settings["project_name"]

    def __init__(self, fmt=None, datefmt=None, style="%"):
        # Do not let mozilla-cloud-services-logger constructor to improperly
        # use style as the logger_name.
        # See https://github.com/mozilla/mozilla-cloud-services-logger/issues/3
        logger_name = self.logger_name
        super().__init__(fmt, datefmt, style)
        self.logger_name = logger_name


def get_user_info(request):
    # Default user info (shown in hello view for example).
    user_info = {"id": request.prefixed_userid, "principals": request.prefixed_principals}
    if hasattr(request, "get_user_profile"):
        user_info["profile"] = request.get_user_profile()
    return user_info


def includeme(config):
    settings = config.get_settings()

    # Heartbeat registry.
    config.registry.heartbeats = {}

    # Public settings registry.
    config.registry.public_settings = {"batch_max_requests", "readonly"}

    # Directive to declare arbitrary API capabilities.
    def add_api_capability(config, identifier, description="", url="", **kw):
        existing = config.registry.api_capabilities.get(identifier)
        if existing:
            error_msg = "The '{}' API capability was already registered ({})."
            raise ValueError(error_msg.format(identifier, existing))

        capability = dict(description=description, url=url, **kw)
        config.registry.api_capabilities[identifier] = capability

    config.add_directive("add_api_capability", add_api_capability)
    config.registry.api_capabilities = {}

    # Resource events helpers.
    config.add_request_method(events.get_resource_events, name="get_resource_events")
    config.add_request_method(events.notify_resource_event, name="notify_resource_event")

    # Setup cornice.
    config.include("cornice")

    # Setup cornice api documentation
    config.include("cornice_swagger")

    # Per-request transaction.
    config.include("pyramid_tm")

    # Add CORS settings to the base kinto.core Service class.
    Service.init_from_settings(settings)

    # Use the project name as the main logger name (Logger field in MozLog).
    JsonLogFormatter.init_from_settings(settings)

    # Setup components.
    for step in aslist(settings["initialization_sequence"]):
        step_func = config.maybe_dotted(step)
        step_func(config)

    # Custom helpers.
    config.add_request_method(log_context)
    config.add_request_method(follow_subrequest)
    config.add_request_method(prefixed_userid, property=True)
    config.add_request_method(prefixed_principals, reify=True)
    config.add_request_method(get_user_info, name="get_user_info")
    config.add_request_method(current_resource_name, reify=True)
    config.add_request_method(current_service, reify=True)
    config.commit()

    # Include plugins after init, unlike pyramid includes.
    includes = aslist(settings["includes"])
    for app in includes:
        config.include(app)

    # # Show settings to output.
    # for key, value in settings.items():
    #     logger.info('Using {} = {}'.format(key, value))

    # Scan views.
    config.scan("kinto.core.views")

    # Give sign of life.
    msg = "Running {project_name} {project_version}."
    logger.info(msg.format_map(settings))



from cornice_swagger import CorniceSwagger
from cornice_swagger.converters.schema import TypeConverter

from kinto.core.schema import Any


class AnyTypeConverter(TypeConverter):
    """Convert type agnostic parameter to swagger."""

    def __call__(self, schema_node):
        return {}


class OpenAPI(CorniceSwagger):
    """OpenAPI documentation generator."""

    custom_type_converters = {Any: AnyTypeConverter}
    """Kinto additional type converters."""

    security_definitions = {}
    """Kinto security definitions. May be used for setting other security methods."""

    security_roles = {}
    """Kinto resource security roles. May be used for setting OAuth roles by plugins."""

    @classmethod
    def expose_authentication_method(cls, method_name, definition):
        """Allow security extensions to expose authentication methods on the
        OpenAPI documentation. The definition field should correspond to a
        valid OpenAPI security definition. Refer the OpenAPI 2.0 specification
        for more information.
        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md

        Below are some examples for BasicAuth and OAuth2::

            {
                "type": "basic",
                "description" "My basicauth method."

            }

            {
                "type": "oauth2",
                "authorizationUrl": "https://oauth-stable.dev.lcip.org/v1",
                "flow": "implicit",
                "scopes": {"kinto": "Kinto user scope."}
            }

        """
        cls.security_definitions[method_name] = definition
        cls.security_roles[method_name] = definition.get("scopes", {}).keys()

    def __init__(self, services, request):
        super().__init__(services)

        self.request = request
        self.settings = request.registry.settings

        self.api_title = self.settings["project_name"]
        self.api_version = self.settings["http_api_version"]
        self.ignore_ctypes = ["application/json-patch+json"]

        # Matches the base routing address - See kinto.core.initialization
        self.base_path = f"/v{self.api_version.split('.')[0]}"

    def generate(self):
        base_spec = {
            "host": self.request.host,
            "schemes": [self.settings.get("http_scheme") or "http"],
            "securityDefinitions": self.security_definitions,
        }

        return super(OpenAPI, self).generate(swagger=base_spec)

    def default_tags(self, service, method):
        """Povides default tags to views."""

        base_tag = service.name.capitalize()
        base_tag = base_tag.replace("-plural", "s")
        base_tag = base_tag.replace("-object", "s")

        return [base_tag]

    def default_op_ids(self, service, method):
        """Povides default operation ids to methods if not defined on view."""

        method = method.lower()
        method_mapping = {"post": "create", "put": "update"}
        if method in method_mapping:
            method = method_mapping[method]

        resource = service.name
        if method == "create":
            resource = resource.replace("-plural", "")

        resource = resource.replace("-plural", "s")
        resource = resource.replace("-object", "")
        op_id = f"{method}_{resource}"

        return op_id

    def default_security(self, service, method):
        """Provides OpenAPI security properties based on kinto policies."""

        definitions = service.definitions

        # Get method view arguments
        for definition in definitions:  # pragma: no branch
            met, view, args = definition
            if met == method:
                break

        if args.get("permission") == "__no_permission_required__":
            return []
        else:
            return [{name: list(roles)} for name, roles in self.security_roles.items()]



import logging
from functools import wraps
from math import ceil, floor
from time import time

from pyramid.settings import aslist

from kinto.core.cache import CacheBase
from kinto.core.storage import exceptions
from kinto.core.utils import json, memcache

logger = logging.getLogger(__name__)


def wrap_memcached_error(func):
    @wraps(func)
    def wrapped(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except TypeError:
            raise
        except (
            memcache.Client.MemcachedKeyError,
            memcache.Client.MemcachedStringEncodingError,
        ) as e:
            logger.exception(e)
            raise exceptions.BackendError(original=e)

    return wrapped


def create_from_config(config, prefix=""):
    """Memcached client instantiation from settings.
    """
    settings = config.get_settings()
    hosts = aslist(settings[prefix + "hosts"])
    return memcache.Client(hosts)


class Cache(CacheBase):
    """Cache backend implementation using Memcached.

    Enable in configuration::

        kinto.cache_backend = kinto.core.cache.memcached

    *(Optional)* Instance location URI can be customized::

        kinto.cache_hosts = 127.0.0.1:11211 127.0.0.1:11212

    :noindex:

    """

    def __init__(self, client, *args, **kwargs):
        super(Cache, self).__init__(*args, **kwargs)
        self._client = client

    def initialize_schema(self, dry_run=False):
        # Nothing to do.
        pass

    @wrap_memcached_error
    def flush(self):
        self._client.flush_all()

    @wrap_memcached_error
    def _get(self, key):
        value = self._client.get(self.prefix + key)
        if not value:
            return None, 0
        data = json.loads(value)
        return data["value"], data["ttl"]

    def ttl(self, key):
        _, ttl = self._get(key)
        val = ttl - time()
        return floor(val)

    def get(self, key):
        value, _ = self._get(key)
        return value

    @wrap_memcached_error
    def expire(self, key, ttl):
        if ttl == 0:
            self.delete(key)
        else:
            # We can't use touch here because we need to update the TTL value in the record.
            value = self.get(key)
            self.set(key, value, ttl)

    @wrap_memcached_error
    def set(self, key, value, ttl):
        if isinstance(value, bytes):
            raise TypeError("a string-like object is required, not 'bytes'")
        value = json.dumps({"value": value, "ttl": ceil(time() + ttl)})
        self._client.set(self.prefix + key, value, int(ttl))

    @wrap_memcached_error
    def delete(self, key):
        value = self.get(key)
        self._client.delete(self.prefix + key)
        return value


def load_from_config(config):
    settings = config.get_settings()
    client = create_from_config(config, prefix="cache_")
    return Cache(client, cache_prefix=settings["cache_prefix"])



import unittest
from unittest import mock

from pyramid import httpexceptions

from kinto.core.resource import Resource, ShareableResource
from kinto.core.storage import exceptions as storage_exceptions
from kinto.core.testing import DummyRequest

from . import BaseTest


class ResourceTest(BaseTest):
    def test_raise_if_backend_fails_to_obtain_timestamp(self):
        request = self.get_request()

        with mock.patch.object(
            request.registry.storage,
            "resource_timestamp",
            side_effect=storage_exceptions.BackendError,
        ):
            with self.assertRaises(storage_exceptions.BackendError):
                self.resource_class(request)

    def test_raise_unavailable_if_fail_to_obtain_timestamp_with_readonly(self):
        request = self.get_request()

        excepted_exc = httpexceptions.HTTPServiceUnavailable

        request.registry.settings = {"readonly": "true"}
        with mock.patch.object(
            request.registry.storage,
            "resource_timestamp",
            side_effect=storage_exceptions.BackendError,
        ):
            with self.assertRaises(excepted_exc) as cm:
                self.resource_class(request)
                self.assertIn("writable", cm.exception.message)

    def test_resource_can_be_created_without_context(self):
        try:
            self.resource_class(self.get_request())
        except Exception as e:
            self.fail(e)

    def test_default_parent_id_is_empty(self):
        request = self.get_request()
        parent_id = self.resource.get_parent_id(request)
        self.assertEqual(parent_id, "")


class DeprecatedShareableResource(unittest.TestCase):
    def test_deprecated_warning(self):
        with mock.patch("warnings.warn") as mocked_warnings:
            ShareableResource(context=mock.MagicMock(), request=mock.MagicMock())

        message = "`ShareableResource` is deprecated, use `Resource` instead."
        mocked_warnings.assert_called_with(message, DeprecationWarning)


class DeprecatedMethodsTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        patch = mock.patch("warnings.warn")
        self.mocked_warnings = patch.start()
        self.addCleanup(patch.stop)

        req = DummyRequest()
        req.validated = {"body": {}, "header": {}, "querystring": {}}
        req.registry.storage.list_all.return_value = []
        req.registry.storage.delete_all.return_value = []
        req.registry.storage.create.return_value = {"id": "abc", "last_modified": 123}

        self.resource = Resource(context=mock.MagicMock(), request=req)

    def test_record_id(self):
        self.resource.record_id

        message = "`record_id` is deprecated, use `object_id` instead."
        self.mocked_warnings.assert_called_with(message, DeprecationWarning)

    def test_process_record(self, *args, **kwargs):
        self.resource.process_record(new={}, old=None)

        message = "`process_record()` is deprecated, use `process_object()` instead."
        self.mocked_warnings.assert_called_with(message, DeprecationWarning)

    def test_collection_get(self, *args, **kwargs):
        self.resource.collection_get()

        message = "`collection_get()` is deprecated, use `plural_get()` instead."
        self.mocked_warnings.assert_called_with(message, DeprecationWarning)

    def test_collection_post(self, *args, **kwargs):
        self.resource.collection_post()

        message = "`collection_post()` is deprecated, use `plural_post()` instead."
        self.mocked_warnings.assert_called_with(message, DeprecationWarning)

    def test_collection_delete(self, *args, **kwargs):
        self.resource.collection_delete()

        message = "`collection_delete()` is deprecated, use `plural_delete()` instead."
        self.mocked_warnings.assert_called_with(message, DeprecationWarning)


class NewResource(Resource):
    def get_parent_id(self, request):
        return "overrided"


class ParentIdOverrideResourceTest(BaseTest):
    resource_class = NewResource

    def test_get_parent_can_be_overridded(self):
        request = self.get_request()

        parent_id = self.resource.get_parent_id(request)
        self.assertEqual(parent_id, "overrided")
        self.assertEqual(self.resource.model.parent_id, "overrided")


class CustomModelResource(Resource):
    def __init__(self, *args, **kwargs):
        self.model = mock.MagicMock()
        self.model.name = mock.sentinel.model
        super().__init__(*args, **kwargs)


class CustomModelResourceTets(unittest.TestCase):
    def test_custom_model_is_not_overriden(self):
        c = CustomModelResource(request=mock.MagicMock())
        self.assertEqual(c.model.name, mock.sentinel.model)
